(** Pierre Castéran, inspired   from Sladek's paper "The Termite and the Tower"
 *)
Require Import ArithRing Omega.
Require Import Epsilon0.Epsilon0 KS.
Require Import omega_iterates.

(* Functional Iteration *)
(*
Fixpoint iterate {A:Type}(f : A -> A) (n: nat)(x:A) :=
  match n with
      0 => x
    | S p => f (iterate  f p x)
  end.
*)
(* Falpha's *)

Import Restriction.
About FixR.
Check FixR (R:= LT)(E:= nf) (P:= fun alpha => nat -> nat).

(*
transfinite_recursor_LT :
forall P : T1 -> Type,
(forall x : T1, nf x -> (forall y : T1, y < x -> P y) -> P x) ->
forall a : T1, nf a -> P a
 *)

Check transfinite_recursor_LT (fun alpha => nat -> nat).
                         
Section build_falpha.
 Variable alpha : T1.
 Hypothesis alpha_nf : nf alpha.
   Variable rec : T1 -> nat -> nat.
 

  About zero_succ_limit_dec.
   
 Definition funalpha :=
     match (zero_succ_limit_dec alpha_nf)
     with
       | inl (left _) => S
       | inl (right Hl) =>
         fun n => rec (dS n alpha) n 
                    
       |  inr (exist _ gamma (conj H1 H2)) =>
          fun n => iterate (rec gamma)  (S n) n
                  
     end.


 
 (*Definition funalpha :=
     match (zero_succ_limit_dec alpha_nf)
     with
       | inl (left _) => S
       | inl (right Hl) =>
         fun n => rec (dS n alpha)
                      (nf_dS n alpha_nf )
                      (dS_lt n alpha_nf
                                 (is_limit_not_zero alpha_nf  Hl)) n
       |  inr (exist _ gamma (conj H1 H2)) =>
          fun n => iterate (rec gamma H1
                       (eq_succ_lt  H1 H2)) (S n) n
     end.

 *)


 End build_falpha.
About funalpha.

Lemma funalpha_zero_rw : forall f i, funalpha zero zero_nf f i = S i.
unfold funalpha;simpl. 
auto. 
Qed.
About funalpha.

Lemma funalpha_succ_rw : forall f alpha i (H: nf alpha),
                           funalpha (succ alpha) (succ_nf H)  f i =
                           iterate (f alpha )  (S i) i.
intros; unfold funalpha;simpl. 
 destruct (zero_succ_limit_dec (succ_nf H)).
 destruct s. 
 SearchAbout (succ _ <> zero).
 destruct (succ_not_zero _ e).
 SearchAbout (is_limit (succ _)).
  destruct  (@is_limit_succ alpha); auto.  
 destruct s.
  destruct a;subst.
  SearchAbout (succ _ = succ _).
  SearchAbout succ.
  specialize (succ_injective H n e). intro; now subst.
  Qed. 



Compute fun f => funalpha zero zero_nf f 42.


(*

Definition F_long  (alpha:T1)(Halpha : nf alpha) :=
  FixR (R:= lt)(E:= nf) (P:= fun alpha => nat -> nat) funalpha nf_Wf alpha Halpha.

About F_long.



About omega_iterates.F_.

Lemma F_zero_eqn (H: nf zero) : forall n, F_long zero  H n = S n.
intros; unfold F_long.  
SearchAbout FixR.
Print FixR.
unfold FixR; cbn.
About  FixR_F_eq. 
generalize (FixR_F_eq T1 nf lt (fun alpha => nat -> nat) funalpha ). 

intro HH.
unfold funalpha in HH.
specialize (HH zero zero_nf).

simpl (zero_succ_limit_dec zero_nf) in HH.
SearchAbout Acc.
specialize (HH (nf_Acc zero_nf)).
Set Printing Implicit.

case_eq (@inl ({zero = zero} + {false = true})
           {beta : T1 | nf beta /\ zero = succ beta}
           (@left (zero = zero) (false = true) (@eq_refl T1 zero))).
intros.
rewrite H0 in HH.
destruct a.
 rewrite  HH.
Admitted. 
 reflexivity. 
destruct ( @inl ({zero = zero} + {false = true})). simpl in HH.
unfold FixR_F in HH.
cbn in HH.
SearchAbout Acc.
cbn. 
Admitted.

(** TO DO : Check consistency with finite ordinals *)


Lemma F_succ_eqn :
  forall beta (Hbeta : nf beta) n, 
    F_long  (succ beta) (succ_nf Hbeta) n =
    Fsuc (F_long beta Hbeta) n.
Admitted.

(* lets try to have alpha = alpha' as an hypothesis *)

Lemma F_long_proof_irrelevance : forall alpha
                                   (Halpha Halpha' : nf alpha)
                                   n,
                              F_long alpha  Halpha n =
                              F_long  alpha Halpha' n.
Admitted.
Search nf.



Lemma L332: 

  forall i, FS (S i) = succ (FS i).
  reflexivity.
Qed.

About nf_finite.

About iterate. 
Lemma iterate_of_extensional {A}(f g: A -> A) (H: forall x, f x = g x):
  forall n x, iterate f n x = iterate g n x.
induction n. 
 simpl; auto. 
 intro x; simpl. 
  transitivity (f (iterate g n x)).
  now rewrite IHn.  
  apply H. 
Qed. 


Lemma F_finite : forall i n, F_long  (FS i) (nf_finite i) n =
                             omega_iterates.F_ (S i) n.
  induction i.
  intros. 
   admit. (* rewrite (F_long_proof_irrelevance _ (nf_finite 1) _ n).
   cbn; now rewrite F_zero_eqn. *)
   intro n.
   About succ_nf.
   simpl.
Print sig.
assert (exist nf (FS (S i))   (nf_finite  (S i)) =
        exist nf (succ (FS i))   (succ_nf (nf_finite i))).
  f_equal. 
  SearchAbout succ_nf.
  admit.

  dependent rewrite H. 
  rewrite F_succ_eqn.
  SearchAbout (F_ (S _) _).
  rewrite F_S_rw.
  unfold Fsuc.  
  clear H.
   apply   iterate_of_extensional. 
   auto. 

Admitted.



(* 
Check F_ omega.



Definition HAck := F_ (phi0 omega).
 *)


(*
Goal forall n,  falpha zero_nf  n = S n.
  unfold falpha.
  unfold transfinite_recursor_LT.
  unfold transfinite_recursor.
  intros.
  Abort.
 *)
 *)

