(** An implementation through lists of "alpha-large" sets 
    (After Ketonen and Solovay's "rapidly growing Ramsey functions" )
*)

(*  Pierre Casteran 
    Univ. Bordeaux, laBRI
*)

(* TO DO :
  Make precise references to KS (Ketonen Solovay) or KP (Kirby Paris) 
  theorems and definitions
 *)

(** Work in Progress !!! 
    We're switching from one notation to another ; thus many proofs are still 
    duplicated
    Proofs are still unreadable !!!! *)

Require Import Epsilon0 KS MoreLists.
Import Relation_Operators Sorted.
(* Coercion fin : nat >-> T1. *)

(** Like alpha<s> but s is a sequence instead of a set.
                       and all elements of s are shifted by 1.
   So, our (gnawS alpha (i::j:: ... ::l::nil)) is just
          alpha <{i+1,j+1, ... l+1}>
 *)


Fixpoint gnawS (s: list nat) (alpha : T1) :=
  match alpha, s with
    | _, nil => alpha
    | _, (i::Y) => gnawS Y (dS i alpha)
  end.

Fixpoint gnaw (s: list nat) (alpha : T1) :=
  match alpha, s with
    | _, nil => alpha
    | _, (0::Y) => gnaw Y alpha
    | _, (S i :: s')  =>  gnaw s' (dS i alpha)
  end.

Example ex1 :  gnaw (interval 1 303) (omega ^ 3  +  2) = 
               omega ^ 2 + omega * 93  + 83.
 reflexivity.
Qed.


Lemma gnaw_gnawS (s:list nat) :
  forall alpha,  gnawS s alpha = gnaw (List.map S s) alpha.
Proof.
 induction s.
  - reflexivity.
  - cbn.  intros. now  rewrite IHs.
 Qed.

Fixpoint preds (l : list nat) : list nat :=
  match l with
      nil => nil
    | 0 :: l' => preds l'
    | S i :: l' => i :: preds l'
  end.

Lemma gnawS_gnaw (s:list nat) :
  forall alpha,  gnaw s alpha = gnawS (preds s) alpha.
Proof.
 induction s.
  - reflexivity.
  - cbn.  intros.
    destruct a. auto. now  rewrite IHs.
Qed. 


Definition largeSb (alpha : T1) (s: list nat) :=
  match gnawS s alpha with
    | zero => true
    | _ => false
  end.

Definition largeb (alpha : T1) (s: list nat) :=
  match gnaw s alpha with
    | zero => true
    | _ => false
  end.


Lemma largeSb_b (alpha : T1) (s: list nat) :
  largeSb alpha s = largeb alpha (List.map S s). 
Proof.
  generalize alpha; induction s;cbn.
  - trivial.
  - cbn. unfold largeb, largeSb.
 cbn. intros; now rewrite <- gnaw_gnawS.
Qed. 


Lemma largeb_Sb alpha s :
  largeb alpha s = largeSb alpha (preds s).
  unfold largeb, largeSb.
  now rewrite  gnawS_gnaw.
Qed.

Definition large (alpha : T1) (s : list nat) : Prop :=
  largeb alpha s.

Definition largeS (alpha : T1) (s : list nat) : Prop :=
  largeSb alpha s.
  
Lemma gnawS_omega : forall i s, gnawS (i::s) omega = gnawS s (FS i).
Proof.
  intros; now cbn.
Qed.
Lemma gnaw_omega i s : gnaw (S i::s) omega = gnaw s (FS i).
Proof.
 reflexivity.
 Qed.


Lemma gnaws_rw i s alpha : gnawS (i::s) alpha = gnawS s (dS i alpha).
Proof. reflexivity. Qed.

Lemma gnaw_rw i s alpha : gnaw (S i::s) alpha = gnaw s (dS i alpha).
Proof. reflexivity. Qed.


Lemma gnawS_app : forall  s s' alpha,
                   gnawS (s ++ s') alpha  = gnawS s' (gnawS s alpha).
Proof. induction s; cbn; auto. Qed. 

Lemma gnaw_app : forall  s s' alpha,
                   gnaw (s ++ s') alpha  = gnaw s' (gnaw s alpha).
Proof. induction s; cbn.
       - trivial.
       - destruct a; auto.
Qed. 


Lemma largeS_iff (alpha : T1) (s : list nat) :
  largeS alpha s <-> gnawS s alpha = zero.
Proof.  
  unfold largeS, largeSb; destruct (gnawS s alpha).
  - split;auto.  
  - split; discriminate.
Qed.

Lemma large_iff (alpha : T1) (s : list nat) :
  large alpha s <-> gnaw s alpha = zero.
Proof.  
  unfold large, largeb; destruct (gnaw s alpha).
  - split;auto.  
  - split; discriminate.
Qed. 


Lemma gnawS_succ: forall alpha  i s, nf alpha -> 
                                    gnawS (i::s) (succ alpha) =
                                    gnawS s alpha.
Proof.   intros; cbn; now rewrite dS_succ. Qed. 


Lemma gnaw_succ: forall alpha  i s, nf alpha -> 
                                    gnaw (S i::s) (succ alpha) =
                                    gnaw s alpha.
Proof.   intros; cbn; now rewrite dS_succ. Qed. 


(* ICI *)


Lemma largeSb_of_succ : forall alpha  i s, nf alpha -> 
                                          largeSb (succ alpha) (i::s) =
                                          largeSb alpha s.
Proof.
  intros; unfold largeSb; now rewrite gnawS_succ. 
Qed.

Lemma largeb_of_succ : forall alpha  i s, nf alpha -> 
                                          largeb (succ alpha) (S i::s) =
                                          largeb alpha s.
Proof.
  intros; unfold largeb; now rewrite gnaw_succ. 
Qed. 





Lemma gnawS_finite : forall n i, gnawS (first_nats_aux i (S n)) (FS n) =
                                zero.
Proof.
  induction n.
  - intro; now cbn. 
  - intro i;  specialize (IHn (S i)).
    rewrite <- succ_compat; simpl first_nats_aux; rewrite gnawS_succ; auto. 
Qed.

Lemma gnaw_finite : forall n i, gnaw (first_nats_aux (S i) (S n)) (FS n) =
                                zero.
Proof.
  induction n.
  - intro; now cbn. 
  - intro i;  specialize (IHn (S i)).
    rewrite <- succ_compat; simpl first_nats_aux; rewrite gnaw_succ; auto. 
Qed.



Lemma largeSb_finite :
  forall n i, largeSb (FS n) (first_nats_aux i (S n)) = true.
Proof.
  intros; unfold largeSb; now rewrite gnawS_finite. 
Qed.


Lemma largeb_finite :
  forall n i, largeb (FS n) (first_nats_aux (S i) (S n)) = true.
Proof.
  intros; unfold largeb; now rewrite gnaw_finite. 
Qed.


Lemma gnawS_finite' :
  forall n i, gnawS (first_nats_aux i n) (FS n) = one.
Proof.
  induction n.
  - intro; now cbn. 
  - intro i;  rewrite <- succ_compat;  simpl first_nats_aux. 
    specialize (IHn  (S i)); rewrite gnawS_succ; auto.
Qed.

Lemma gnaw_finite' :
  forall n i, gnaw (first_nats_aux (S i) n) (FS n) = one.
Proof.
  induction n.
  - intro; now cbn. 
  - intro i;  rewrite <- succ_compat;  simpl first_nats_aux. 
    specialize (IHn  (S i)); rewrite gnaw_succ; auto.
Qed.

Lemma gnawS_omega_n_SSn :
  forall n, gnawS (first_nats_aux n (S (S n))) omega = zero.
Proof. 
  intro n;  destruct n.  
  -  reflexivity. 
  -  apply gnawS_finite.
Qed.

Lemma gnaw_omega_n_SSn :
  forall n, gnaw (first_nats_aux (S n) (S (S n))) omega = zero.
Proof. 
  intro n;  destruct n.  
  -  reflexivity. 
  -  apply gnaw_finite.
Qed.


Lemma gnawS_omega' :
  forall n, gnawS (first_nats_aux  n (S n)) omega = one.
Proof. 
  intro n; destruct n.  
  -  reflexivity. 
  -  apply gnawS_finite'.
Qed.

Lemma gnaw_omega' :
  forall n, gnaw (first_nats_aux  (S n) (S n)) omega = one.
Proof. 
  intro n; destruct n.  
  -  reflexivity. 
  -  apply gnaw_finite'.
Qed.

Lemma gnawS_zero : forall s, gnawS s zero = zero.
Proof. 
  induction s; now cbn.
Qed.

Lemma gnaw_zero : forall s, gnaw s zero = zero.
Proof. 
  induction s; try cbn; auto.
  now destruct a.
Qed.

Lemma gnawS_nf : forall s alpha, nf alpha -> nf (gnawS s alpha).
Proof.
  induction s.
  - intros; now cbn.
  - intros; cbn; apply IHs;  apply nf_dS;auto. 
Qed.


Lemma gnaw_nf : forall s alpha, nf alpha -> nf (gnaw s alpha).
Proof.
  induction s.
  - intros; now cbn.
  - intros; cbn.
    destruct a.
    + auto.
    + apply IHs;  apply nf_dS;auto. 
Qed.


Definition bounded_by (n:nat)(s: list nat) :=
  List.Forall (fun i => i<=n)%nat s.


Definition dS_rel_bounded_by (n:nat) alpha beta :=
  exists i:nat, (i <= n)%nat /\ beta = dS i alpha. 

Definition d_rel_bounded_by (n:nat) alpha beta :=
  exists i:nat, (i <= n)%nat /\ beta = d (S i) alpha. 


(** alpha ---> beta in KP 
             n
 *)

Print d.

Lemma d_dS : forall alpha i, dS i alpha = d (S i) alpha. 
Proof.
  reflexivity.
Qed.



Definition KP_arrow n := clos_trans_1n T1 (dS_rel_bounded_by n).


Lemma KP_5_iii : forall n alpha beta, nf alpha ->  nf beta -> beta <> zero ->
                                      dS_plus n alpha beta <->
                                      KP_arrow n alpha beta.
  split.
  - induction 1.
    + red in H2; left; red.
      exists n; split; auto with arith.
    +  right with y.
       * exists n;   split;auto.
         * apply IHclos_trans_1n;auto;   red in H2;  subst y;  now apply nf_dS.
  - induction 1.
   +  destruct H2; destruct H2.
      apply C12' with x0;auto.
      * subst y; apply dS_LT; auto.
        intro; subst x.    
        destruct H1;now rewrite dS_zero.
      *  left; red;auto.
    +   apply dS_plus_trans with y.
       *  destruct H2; apply C12' with x0;auto.
          destruct H2;  subst y; apply dS_LT; auto.
          intro; subst x.  
          rewrite dS_zero in *.
          Import Operators_Properties.
          clear IHclos_trans_1n; destruct H1.
          { clear H0 x0 H2 H.
            rewrite <- clos_trans_t1n_iff in H3;
              rewrite clos_trans_tn1_iff in H3.
            induction H3.
            destruct H. destruct H. subst.
            now cbn.
            rewrite IHclos_trans_n1 in H.
            destruct H;    destruct H;  subst z; now cbn.
          }   
          tauto.
          destruct H2;left;auto.
     *  apply IHclos_trans_1n;auto.
        destruct H2; destruct H2.
        subst y; apply dS_LT; auto.
        apply nf_dS; auto.
        intro H4; rewrite H4 in H3.
        destruct H1.
        { clear H0 x0 H2 IHclos_trans_1n H4. 
          rewrite <- clos_trans_t1n_iff in H3;
            rewrite clos_trans_tn1_iff in H3.
          induction H3.
          destruct H0. destruct H0. subst.
          now cbn.
          rewrite IHclos_trans_n1 in H0.
          destruct H0;  destruct H0.
          subst z; now cbn.
        }   
Qed.


(** Proposition KS-4.2.a *)

Lemma largeSb_n (n:nat): forall s,  sorted_ge 0 s ->
                                    largeSb (fin n) s = true ->
                                    (n <= List.length s)%nat.
Proof.   
  induction n.
  -  intros; cbn; auto with arith. 
  -  destruct s; cbn.
     +   discriminate. 
     +  intro;   unfold largeSb; cbn.
        destruct n.
      *   replace (gnawS s zero) with zero.      
          auto with arith.
          now rewrite gnawS_zero.
      *   intros;  apply le_n_S, IHn.
          eapply sorted_tail; eauto.
        auto. 
Qed. 


Lemma largeb_n (n:nat): forall s,  sorted_ge 1 s ->
                                    largeb (fin n) s = true ->
                                    (n <= List.length s)%nat.
Proof.   
  induction n.
  -  intros; cbn; auto with arith. 
  -  destruct s; cbn.
     +   discriminate. 
     +  destruct n0.
        *  intros H _; destruct (sorted_inv_gt _ _ _ H); auto with arith. 
        * unfold largeb. intro; cbn. 
        destruct n.
         replace (gnaw s zero) with zero.      
          auto with arith.
          now rewrite gnaw_zero.
         intros;  apply le_n_S, IHn.
          eapply sorted_tail; eauto.
        auto. 
Qed. 



Lemma gnawS_n_R : forall s n,  sorted_ge 0 s -> 
                              (n < List.length s)%nat ->
                              gnawS s (FS n) = zero.
Proof.
  induction s.
  -   cbn;  intros;  destruct n.
      +   cbn; auto. 
          inversion H0.
      +   cbn;  inversion H0. 
  -   cbn;  destruct n. 
      +   cbn; auto.
          now rewrite gnawS_zero. 
      +   intros; apply IHs.
          eapply sorted_tail; eauto.         
          auto with arith. 
Qed.

Lemma gnaw_n_R : forall s n,  sorted_ge 1 s -> 
                             (n < List.length s)%nat ->
                             gnaw s (FS n) = zero.
Proof.
  induction s.
  -   cbn;  intros;  destruct n.
      inversion H0.
      inversion H0.
  - 
    cbn;  destruct n. 
    +   cbn; auto.
        destruct a.
        intro H; destruct (sorted_inv_gt _ _ _ H);auto with arith.
        now rewrite gnaw_zero. 
    +   intros.
        destruct a.
        destruct (sorted_inv_gt _ _ _ H);auto with arith.
        apply IHs.
        eapply sorted_tail; eauto.         
        auto with arith. 
Qed.




Lemma largeSb_n_R : forall s n,  sorted_ge 0 s -> 
                                (n < List.length s)%nat ->
                                largeSb (FS n) s  = true.
Proof.   
  intros; unfold largeSb;now rewrite gnawS_n_R.
Qed.

Lemma largeb_n_R : forall s n,  sorted_ge 1 s -> 
                                (n < List.length s)%nat ->
                                largeb (FS n) s  = true.
Proof.   
  intros; unfold largeb;now rewrite gnaw_n_R.
Qed.

Lemma large_n_iff : forall s (n:nat) , sorted_ge 1 s ->
                                large n  s <-> (n <= List.length s)%nat.
Proof.
  split.
  -  now apply largeb_n.
  -  destruct n.
     + unfold large, largeb; cbn; now rewrite gnaw_zero.
     + unfold large; now apply largeb_n_R.
 Qed. 


Lemma largeS_omega_1 : forall s n,  sorted_ge 0 (n:: s)
                                    -> gnawS (n::s) omega  = zero ->
                                    ( n < List.length s)%nat.
Proof. 
  intros; apply largeSb_n.
  -   eapply sorted_tail; eauto. 
  -    cbn in *;  unfold largeSb;  now rewrite H0.
Qed. 


Lemma large_omega_1 : forall s n,  sorted_ge 1 (n:: s)
                                   -> gnaw (n::s) omega  = zero ->
                                   ( n <= List.length s)%nat.
Proof. 
  intros.
  destruct n.
  - auto with arith. 
  - 
    cbn in H0.
    generalize (largeb_n (S n) s).  
    intros.
    apply H1;auto.
    eapply sorted_le;eauto.
    eapply sorted_tail; eauto.  
    simpl fin.
    unfold largeb.
    now rewrite H0.
Qed. 



Lemma largeS_omega_2 : forall s n,  sorted_ge 0 (n:: s) ->
                                     ( n < List.length s)%nat ->
                                     gnawS (n::s) omega  = zero.
Proof. 
  intros;  cbn;   apply gnawS_n_R; auto. 
  eapply sorted_tail; eauto.
Qed.


Lemma large_omega_2 : forall s n,  sorted_ge 1 (n:: s) ->
                                     ( n <=  List.length s)%nat ->
                                     gnaw (n::s) omega  = zero.
Proof. 
  intros;  cbn;  destruct n.
  destruct (sorted_inv_gt _ _ _ H);auto with arith.
  apply gnaw_n_R; auto. 
  eapply sorted_tail; eauto.
Qed.

Lemma large_omega_iff : forall s n,  sorted_ge 1 (n:: s) ->
                                     (n <=  List.length s)%nat <->
                                     large omega (n::s).
Proof.
 intros s n H; split; rewrite large_iff; intro H0.
 - now apply large_omega_2.  
 - now  apply large_omega_1.
Qed.




Lemma Lemma_4_4_0 :
  forall n p alpha beta (Ha' : alpha <> zero),
    nf alpha -> nf beta -> (n <= p)%nat ->
    dS_star n alpha beta ->
    dS_star p (dS p alpha) (dS p beta).
Proof.
  intros; destruct H2.
  - left;  apply (C12'  H (n:= p)) in H2; auto.
    destruct H2.
    +   red in H2;  subst y;  left;  now red. 
    +  red in H2;  subst y;  apply dS_plus_trans with  z;  auto. 
       left;red;auto. 
    +    eapply dS_plus_LT; eauto.
   -     subst beta; now right.
Qed.


Lemma Lemma_4_4_1 : forall s n alpha beta (Ha : nf alpha)
                           (Hb : nf beta),
                      sorted_ge n s ->
                      dS_star n alpha beta ->
                      dS_star (simple_last s n)
                              (gnawS s alpha) (gnawS s beta).
Proof.
  induction s.
  - intros; now cbn.
  - intros;  cbn;  apply IHs.
   +     apply nf_dS;eauto.
   +   apply nf_dS;eauto.
   +  eapply    sorted_tail'; eauto. 
   +     destruct (T1_eq_dec alpha zero). 
    *    subst;  assert (beta=zero).
         { destruct H0.
           - now apply dS_plus_zero in H0.
           - now symmetry.
         }
    subst; now right.
    * eapply Lemma4_3 with  (n0 :=a);auto. 
      apply C12'' with n; auto. 
      inversion H0; auto with arith.
     all: eapply sorted_head; eauto.
Qed. 


Lemma dS_star_zero:
  forall (n : nat) (alpha : T1), dS_star n zero alpha -> alpha = zero.
Proof. 
  destruct 1.    
  - now apply dS_plus_zero in H.
  - auto.
Qed. 

Section Lemma_4_4_Proof.
  Variables alpha beta : T1.
  Hypothesis Halpha : nf alpha.
  Hypothesis Hbeta : nf beta.
  Variable n : nat.
  Variable s : list nat.
  
  Hypothesis Hs : sorted_ge n s.
  Hypothesis H :  dS_star n alpha beta.
  Hypothesis H0 : largeS alpha s.

  Lemma Lemma4_4 : largeS beta s.
  Proof.
     specialize (Lemma_4_4_1 s n alpha beta Halpha Hbeta Hs H);  intro H1.
     repeat red in H0; unfold largeSb in H0; destruct (gnawS s alpha).
     -   apply dS_star_zero in H1;  now rewrite largeS_iff. 
     -  discriminate. 
  Qed.

End Lemma_4_4_Proof.

Lemma Lemma_4_4_2 : (* K-P-6 *)
  forall (s : list nat) (n : nat) (alpha beta : T1),
    nf alpha -> nf beta ->   sorted_ge n s ->
    dS_star n alpha beta ->  (gnawS s beta) <= (gnawS s alpha).
Proof.
  intros;  specialize (Lemma_4_4_1 s n alpha beta H H0 H1 H2).
  intro H4;  eapply dS_star_LE_2; eauto.                              
  all: eapply gnawS_nf; eauto.
Qed.

    
Lemma Lemma_4_5_1 :  forall n alpha, nf alpha ->
                                     forall s t , ptwise_le s t ->
                                               sorted_ge n s ->
                                               sorted_ge n t ->
                                               dS_star (simple_last t n)
                                                       (gnawS t  alpha)
                                                       (gnawS s  alpha).
Proof. 
  intros n alpha Halpha s t H; unfold ptwise_le in H.
  pattern s , t; eapply  Forall2_indR with Peano.le; eauto.
  - intros; cbn; now right.
  -  intros; subst; repeat rewrite gnawS_app.
     replace (simple_last (l' ++ y :: nil) n) with y.
     assert (dS_star y (gnawS  l' alpha) (gnawS l alpha)).
     {
       eapply C12'' with (simple_last l' n); eauto.
       eapply gnawS_nf; auto.
       eapply sorted_cut;eauto. 
       eapply H1.
       all : eapply sorted_ge_prefix; eauto. 
     }
     eapply  Lemma4_3 with (simple_last l' n); auto.
     eapply gnawS_nf; eauto.
     eapply gnawS_nf; eauto.
     eapply sorted_cut;eauto. 
     eapply H1.
     eapply sorted_ge_prefix; eauto. 
     eapply sorted_ge_prefix; eauto. 
     now rewrite simple_last_app1.
Qed.



Section Proof_of_4_5_2.
 Variables (A B1 B2 : list nat).
 Variable alpha : T1.
 Hypothesis Halpha : nf alpha.
 Hypothesis HA : sorted_ge 0 A.
 Hypothesis HB : sorted_ge 0 (B1 ++ B2).
 Hypothesis HAB1 : ptwise_le B1 A.
 Hypothesis HlargeA : largeS alpha  A .

 Remark R1 : gnawS A alpha = zero.
 Proof. now   rewrite <- largeS_iff. Qed.

 Remark R2 : gnawS  B1 alpha = zero.
   assert (H: sorted_ge 0 B1) by 
    (eapply sorted_ge_prefix; eauto).
   specialize (Lemma_4_5_1 0 alpha Halpha _ _ HAB1 H HA).  
   rewrite R1.
   intro H0; now rewrite (dS_star_zero _ _ H0).
 Qed. 

 Lemma Lemma_4_5_2: gnawS (B1 ++ B2) alpha = zero.
 Proof. 
 rewrite gnawS_app, R2; now rewrite gnawS_zero.
 Qed. 

End Proof_of_4_5_2.



Theorem Theorem_4_5 (alpha: T1)(Halpha : nf alpha)
        (A B : list nat)
        (HA : Sorted Peano.lt A)
        (HB : Sorted Peano.lt B)
        (HAB : incl A B) :
  largeS alpha A -> largeS alpha B.
Proof.
 intro H; destruct (incl_decomp A B HA HB HAB) as [C [D [H1 H2]]];auto.
 subst B; rewrite largeS_iff in *; eapply Lemma_4_5_2 with A;auto.
 rewrite  sorted_ge_iff;  split;auto.
 clear HA H H2 HB HAB; induction A; constructor;auto with arith.
 rewrite  sorted_ge_iff; split; auto.
 generalize (C++D); intro E; induction E; constructor;auto with arith.
 now rewrite largeS_iff in *.
Qed.

Lemma gnawS_lim1 (i:nat)(s: list nat) (lambda : T1) :
  nf lambda -> is_limit lambda ->
  gnawS (i::s) (ocons lambda 0 zero) = gnawS s (ocons (dS i lambda) 0 zero).
Proof. intros;  rewrite gnaws_rw,  dS_lim1; auto.
Qed.


Lemma gnawS_lim2 (i n:nat)(s: list nat) (lambda : T1) :
  nf lambda -> is_limit  lambda->
  gnawS (i::s) (ocons lambda (S n) zero) =
  gnawS s (ocons lambda n (ocons (dS i lambda) 0 zero)).
Proof.
  intros;  rewrite gnaws_rw,  dS_lim2; auto.
Qed.



Lemma gnawS_succ_eqn1 i s gamma :
  nf gamma -> gnawS (i::s) (ocons (succ gamma) 0 zero) = gnawS s (ocons gamma i zero).
Proof.
intros;  rewrite gnaws_rw,  dS_succ_eqn1; auto.
Qed.


Lemma gnawS_succ_eqn2 :
  forall i n s gamma, nf gamma -> 
                      gnawS (i::s) (ocons (succ gamma) (S n) zero) =
                      gnawS  s (ocons  (succ gamma) n  (ocons gamma i zero)).
Proof.
  intros; rewrite gnaws_rw, dS_succ_eqn2;  auto.
Qed.


Lemma gnawS_tail :
  forall i  s alpha n beta,
                  nf (ocons alpha n beta) ->
                  beta <> zero -> 
                  gnawS (i::s) (ocons alpha n beta) =
                  gnawS s (ocons alpha n (dS i beta)).
Proof.
  intros; rewrite gnaws_rw, dS_tail;  auto.
Qed.


Lemma gnawS_SSn (i:nat) s :
  forall alpha n  ,
    nf alpha -> 
    gnawS (i::s)  (ocons alpha (S n) zero) =
    gnawS s (ocons alpha n (dS i (ocons alpha 0 zero))).
Proof.
  intros; rewrite gnaws_rw, dSSn;  auto.
Qed.

Lemma gnawS_cut1 : forall s alpha n beta,
                    nf (ocons alpha n beta) ->
                    gnawS s (ocons alpha n beta) = zero ->
                    exists s1 s2, s = s1 ++ s2 /\
                                  gnawS s1 beta = zero /\
                                  gnawS s2 (ocons alpha n zero) = zero.
Proof.
  induction s.
  -  intros.
     discriminate H0.  
  -   intros; assert (nf alpha) by eauto with T1.
      assert (nf beta) by eauto with T1.
      destruct (T1_eq_dec beta zero).  
      + subst beta;  exists nil, (a::s); repeat split;auto.
      + rewrite gnaws_rw, dS_tail  in H0; auto.
        destruct (IHs alpha n (dS a beta)); auto.
        * apply nf_intro; auto.
          apply nf_dS;auto.
          apply lt_phi0_phi0R.
          apply Epsilon0.lt_trans with beta.
          apply dS_lt;auto.
          apply lt_phi0_phi0.
          eapply lt_phi0_intro; eauto.

        * destruct H3 as [s2 [H4 [H5 H6]]];  exists (a::x), s2.
         subst s; split;auto.
Qed.

Lemma gnawS_cut2 : forall s alpha n ,
                   nf alpha -> 
                   gnawS s (ocons alpha (S n) zero) = zero ->
                   exists s1 s2, s = s1 ++ s2 /\
                                 gnawS s1 (phi0 alpha) = zero /\
                                 gnawS s2 (ocons alpha n zero) = zero.
Proof.
  destruct s. 
  - intros.
    discriminate H0.  
  -  intros.    
     rewrite gnaws_rw in H0.     
     rewrite dSSn in H0; auto.
     destruct (T1_eq_dec alpha zero).
     + subst; exists (n::nil), s;  split;auto.
     + destruct (gnawS_cut1 s alpha n0 (dS n (phi0 alpha))); auto.
       * rewrite <- dSSn; auto.
         apply nf_dS;auto.
       * destruct H1 as [s2 [H2 [H3 H4]]];exists (n::x), s2;  split;auto.
         now subst.
Qed. 

(** Some examples of computation *)

(**  4 < {1,2,3,4} > = 0 *)


Compute gnaw (iota  1) omega.
Compute gnaw (iota  2) omega.

Example omega_1_2_large : large omega (iota 2).
Proof. reflexivity. Qed.

Compute gnaw (interval 10 19) omega.



Goal ~large 42 (interval 100 139).
Proof.    discriminate. Qed.

Goal large 42 (interval 100 141).
Proof. reflexivity.   Qed.

Example omega_10_19_not_large : ~ large omega (interval 10 19).
Proof. discriminate. Qed.


Compute gnaw (interval 1 2) omega.

Compute gnaw (interval 2 3) omega.


Compute gnaw (interval 1 5) (omega * omega).

Compute gnaw (interval 1 12) (omega * omega * omega).

Compute gnaw (interval 1 2) (omega ^ omega).

Compute gnaw (interval 10 122) (omega ^omega).



Example E_4_4 : gnawS (first_nats 4) (fin 4) =  zero.
Proof. reflexivity. Qed.

Example E_3_4 : gnawS (first_nats 3) (fin 4) =  One.
Proof. reflexivity. Qed.

Example E_100_4 : gnawS (first_nats_aux 100 4) (fin 4) = zero.
Proof. reflexivity. Qed.

Example E_100_3 : gnawS (first_nats_aux 100 3) (fin 4) = One.
Proof. reflexivity. Qed.

Example E_3_4_omega : gnawS (interval 3 4) omega = fin 3.
Proof. reflexivity. Qed.

Lemma E_3_4_omega2 :
    gnawS (interval 3 4) (omega* omega) = omega * fin 3 + fin 5.
Proof. reflexivity.  Qed.

Lemma E_2_53_omega2 : gnawS (interval 2 53) (omega ^ omega) =
                      omega * omega * fin 3 + omega * fin 2.
Proof.  reflexivity. Qed.
  
Lemma E_2_54_omega2 : gnawS (interval 2 54) (omega ^ omega) =
omega * omega * fin 3 +  omega + fin 55.
Proof.  reflexivity. Qed.



Example E1 : largeS omega (first_nats_aux 2 4).
Proof. reflexivity. Qed.


Example E2 : largeS (omega * fin 2)  (0::1::2::3::4::5::nil).
Proof. reflexivity. Qed.



Example E3 : gnawS (first_nats 28)  (omega * fin 4) = fin 2.
Proof. reflexivity. Qed.


Example E4 : gnawS (first_nats 200)  (omega * omega * omega + fin 1 ) =
             omega * fin 36 + fin 126.
Proof. reflexivity. Qed. 


Compute gnawS (interval 10 188) (omega + omega + omega + omega)%cnf .


Compute gnawS (interval 10 188) (omega * fin 5)%cnf .

Compute gnawS (interval 10 188) (omega * fin 6)%cnf .

Compute gnawS (interval 10 388) (omega * fin 6)%cnf .

Compute gnawS (interval 10 764) (omega * fin 6)%cnf .

Compute gnawS (interval 10 764) (omega * omega)%cnf .


Compute gnaw (interval 11 765) (omega * omega)%cnf .


Compute gnaw (interval 3  557) (omega ^ omega)%cnf .

Compute gnaw  (iota 40) 42.
(* FS 1 *)

Compute gnaw (iota 41) 42.
(* One *)

Compute gnaw (iota 55) 42.
(* zero *)

Goal  gnaw (interval 100 139) 42 = 2.
reflexivity.
Qed.

Compute gnaw (interval 10 20) 11.

Compute gnaw (interval 100 109) 11.

Compute gnaw (interval 1000 1009) 11.

Compute gnaw (interval 100 109) omega.

Compute gnaw (interval 300 599) omega.

Compute gnaw (interval 1 2) omega.  (* zero *)

Compute gnaw (interval 10 19) omega. (* one *)

Compute gnaw (interval 100 199) omega. (* one *)

Definition test_omega i j := gnaw (interval i j) omega = fin (i - (j -i)).

Goal test_omega 100 145.
reflexivity.
Qed.


Goal test_omega 200 399.
compute.
reflexivity.
Qed.

Definition test_omega2 n := largeb omega (interval n (2 * n)).
Compute test_omega2 44.
Compute test_omega2 123.
Definition test_omega3 n := ~ largeb omega (interval n (Nat.pred (2 * n))).
Compute test_omega3 123.

Compute gnaw (interval 10 41) (omega + omega). (* 1 *)
Compute gnaw (interval 100 401) (omega + omega). (* 1 *)
Compute gnaw (interval 100 200) (omega + omega). (* omega *)
Compute gnaw (interval 100 805) (omega * 3).
Compute gnaw (interval 20 (8 * 20 + 5)) (omega * 3).  (* 1 *)
Compute gnaw (interval 20 (32 * 20 + 29)) (omega * 5).  (* 1 *)

Fixpoint exp2 (i:nat) :=
  match i with
      0 => 1
    | S j => (2 * exp2 j)%nat
  end.

Definition my_test (i j : nat)  :=
     gnaw (interval  j (exp2 i * (j + 1) -3)%nat) (omega * i)
                         = 1.

Compute my_test 4 20.

Compute gnaw (interval 2 14) (omega * omega).

Compute gnaw (interval 3 38) (omega * omega).

Compute gnaw (interval 4 93) (omega * omega).

Compute gnaw (interval 2 4444) (omega * omega * 2).
(*  = ocons One 5 (FS 4257)
     : T1 *)

Example Ex :  gnaw (interval 2 (4444 + 4258)) (omega * omega * 2) = omega * 6.
Proof.  reflexivity. Qed.


(* one *)
(* 1 *)
Compute omega.


(* *)