Require Import Hydra_Games Wellfounded WfVariant Ensembles
        Relations Relation_Operators RelationClasses.

(** A class of proofs that a given  well-founded order is not sufficient 
    to prove the termination of all hydra battles *)


Section S_given.

  (** Let us consider some strict order lt over a type t *)
  
  Variables (t: Type)(lt : relation t).
  Context (S: StrictOrder lt).
  Infix "<" := lt.
  
  Let le := clos_refl _ lt.
  Infix "<=" := le.
  

  (** A class of proofs by reduction ad absurdum 
   *)
  
  Class TooSimple :=
    {
      iota : t -> Hydra;
      too_simple_1 :  forall m (V: WfVariant round lt m) alpha,
                        alpha <= m (iota alpha);
      h: Hydra;                            
      too_simple_2 : forall m (V: WfVariant round lt m),
                       h -+-> iota (m h)
    }.

  Lemma impossibility (too_simple  : TooSimple) :
    forall m (V: WfVariant round lt m), False.
  Proof.
    intros m V;  destruct too_simple  as [iota H0 h H1].
    specialize (H0 m V); specialize (H1 m V).
    destruct V as [H2 H3].
    assert (lt (m (iota (m h)) ) (m h)).
    {
      induction H1.
     - auto.
     - transitivity (m y); auto. 
      }
    specialize (H0 (m h)).
    destruct H0  as [y H4 |].
    -  destruct (@StrictOrder_Irreflexive _ lt S y).
       transitivity (m h);auto.
    - now  destruct (@StrictOrder_Irreflexive _ lt S (m h)).
  Qed.

  
End S_given.



