(* Pierre Casteran ,LaBRI  *)


(** * Partial functions *)
(** * We study the relationship between three representations of partial 
   functions : relational presentation, using the  description operator *)


Require Import Ensembles  ClassicalDescription MoreIota.

Set Implicit Arguments.
Arguments In {U} _ _.

Definition iota_fun {A B:Type}(Hb:inhabited B)(DA: Ensemble A)
           (DB : Ensemble B)(R:A->B-> Prop) :=
  fun a =>  iota Hb (fun b':B => In DA a /\ R a b'  /\ In DB b').

Lemma iota_fun_e {A B:Type}(Hb:inhabited B)(DA: Ensemble A)
      (DB:Ensemble B)(R:A->B-> Prop): 
  forall (a:A), DA a -> 
   (exists ! b, R a b /\ In DB b) ->
   unique (fun b => (R a b /\ In DB b)) (iota_fun Hb DA DB R a).
Proof.
 intros; unfold iota_fun; split.
 - apply iota_ind.
   destruct H0 as [y Hy]; exists y.
   destruct Hy;split.
   + tauto.
   + destruct 1;  auto.
   + destruct 1;  tauto.
 -  destruct 1.
    symmetry;apply iota_eq.
    + exists x';  split.
      tauto.
      destruct 1.
      destruct H0.
      destruct H0.
      symmetry;auto.
      symmetry;auto.
      transitivity x.
      symmetry;apply H5;auto.
      apply H5;auto.
   + tauto.
Qed.




Lemma iota_fun_ind {A B:Type}(Hb:inhabited B)(DA: Ensemble A)(DB:Ensemble B)(P:A -> Prop)
      (Q R:A->B-> Prop):
  forall (a x:A), DA a -> a = x ->
   (exists ! b, R a b /\ In DB b) ->
   P a ->
   (forall b,  unique (fun b => R a b /\  In DB b) b -> Q a b) ->
   Q x (iota_fun Hb P DB R a).
Proof.
 intros.
 subst x.
 apply H3. 
 
generalize (iota_fun_e Hb DA DB R).
  intros.
specialize (H0 a H H1).
 
apply iota_fun_e.
 auto.
 auto.
Qed.



Lemma iota_fun_rw  {A B:Type}: forall Hb:  inhabited B,
  forall (P : A->Prop)(DA: Ensemble A)(DB:Ensemble B)(R:A->B-> Prop)(a:A) (Ha : In DA a)(b':B),
   (exists ! b, R a b /\ In DB b) -> 
   P a -> 
   R a b' -> In DB b' -> b'= (iota_fun Hb P DB R a).
Proof.
 intros.
 unfold iota_fun.
 apply iota_eq. 
 exists b';split;auto.
 destruct 1.
 destruct H.
  transitivity x;auto.
  symmetry;auto.
 destruct H.
 apply H5.
   split;auto.
  destruct H.
 apply H5.
 auto.
 tauto.
Qed.


Goal inhabited nat.
split. exact 42.
Save inh_nat.


Example my_pred := iota_fun inh_nat (fun i=> i> 0) (fun i => True) ( fun x y => x=S y).


Ltac iota_dev := 
 match goal with |- context boum [iota ?x ?p] =>
                  generalize (iota_spec x p) end.

Lemma pred_of_succ : forall n:nat, my_pred (S n) = n.
Proof.
  intro n.
  -   unfold my_pred, iota_fun.
      iota_dev.   
      destruct 1.
     +   exists n;split.
         split;auto.
         red;auto with arith.
         split;auto.
        red;trivial.
intros x' [H1 H2].
 destruct H2.
         injection H;auto.
     + destruct H0.
       symmetry;apply iota_eq.
       exists n;split.
      split;auto.
     destruct 1.
  destruct H3.     injection H3.
    auto.
 repeat split;auto.
Qed.   



Goal my_pred 43 = 42.
apply pred_of_succ.
Qed.


Section f_fixed.

  Variables A B : Type.
  Variable inhA :  inhabited A.
  Variable inhB :  inhabited B.
  (** domain *)
  Variable DA : Ensemble A. 
  (** codomain *)
  Variable DB :  Ensemble B. 
  
  Variable f : A -> B.

  (** relational representation *)
  Variable Rf : A -> B -> Prop. 

  (** equivalence between a function and a relation *)
  Definition fun_rel_same :=
    forall a b, In DA a -> (Rf a b  <-> f a = b).
 
  
  (* abstract properties of a function (relational representation ) *)
  Definition rel_domain := forall a, In DA a -> exists b, Rf a b .
  Definition rel_codomain := forall a b, In DA a -> Rf a b -> In DB b.
  Definition rel_functional := forall a b b', In DA a -> 
                              Rf a b -> Rf a b' -> b = b'.
  Definition rel_onto := forall b, In DB b -> exists a, In DA a /\ Rf a b.
  Definition rel_inj := forall a a' b, In DA a -> 
                                       In DA a' ->
                                       Rf a b -> 
                                       Rf a' b -> 
                                       a = a'.  

  Inductive  rel_injection : Prop :=
   rel_inj_i : rel_domain ->  rel_codomain -> 
        rel_functional -> 
        rel_inj -> 
        rel_injection.

  Inductive  rel_bijection : Prop :=
   rel_bij_i : rel_domain ->  rel_codomain -> 
        rel_functional -> 
        rel_onto ->
        rel_inj -> 
        rel_bijection.



  (** Abstract properties of f: A->B (wrt. domain DA and codomain DB *)

  Definition fun_codomain := forall a, In DA a ->  In DB (f a).
  Definition fun_onto := forall b, In DB b -> exists a, In DA a /\ f a = b.
  Definition fun_inj := forall a a' , In DA a -> In DA a' -> f a = f a' -> 
                 a = a'.  

  Definition image := fun b => exists a, In DA a /\ f a = b.


  Inductive  fun_injection : Prop :=
   fun_inj_i :  fun_codomain -> 
                            fun_inj -> fun_injection.

  Inductive  fun_bijection : Prop :=
   fun_bij_i :  fun_codomain -> fun_onto ->
                            fun_inj -> fun_bijection.
Hint Unfold In.


  Lemma bijection_fun_rel : fun_rel_same -> fun_bijection -> rel_bijection.
  Proof.
    intro Sa; red in Sa.
  destruct 1;split.
   
    red.
  intros; exists (f a).
  case (Sa a (f a) H2). auto.
  red.
  intros.
case (Sa a b H2).
red in H.
 intros H4 _;generalize (H4 H3).
 intro; subst b;auto.
 red;intros.
 case (Sa a b H2);intros.
 case (Sa a b' H2);intros.
 intuition.
 subst b;auto.
 red;intros.
 red in H0.
 case (H0 _ H2).
 intros a (Ha,Ea).
 case (Sa a b Ha).
 intros.
 exists a;auto.
 red;intros.
 red in H1.
 generalize (H1 _ _ H2 H3).
 intro.
 apply H6.
 case (Sa a b H2);case (Sa a' b H3).
 intuition.
 subst b;auto.
Qed.


  Definition rel_inv := fun b a => In DA a /\ In DB b /\ Rf a b.
 
  Lemma bijection_rel_fun : fun_rel_same -> rel_bijection -> fun_bijection.
  Proof.
    intro Sa; red in Sa.
  destruct 1;split.
   
    red.
  red in H0.
 intros ;apply H0 with a; auto.
   case (Sa a (f a) H4).
  auto.
  red.
  intros.
  red in H2.
 case (H2 _ H4).
 intros a (Ha,Ra).
 exists a.
 case (Sa a b Ha).
 auto.
 red.
 red in H3.
 intros.
 apply H3 with (f a);auto.
 case (Sa a (f a) H4);auto.
 rewrite H6; case (Sa a' (f a') H5);auto.
Qed.

 

End f_fixed.

Lemma rel_inv_bij : forall (A B : Type)(DA : Ensemble A)
                                       (DB : Ensemble B)
                   (Rf : A -> B -> Prop),
                   (rel_bijection DA DB Rf) ->
                   (rel_bijection DB DA (rel_inv DA DB Rf)).
 destruct 1; split.
 red.
 intros b Hb.  case (H2  b Hb).
 intros x (Inx,Rxb). exists x;red;auto.
 red.
 intros b a Inb Rba.
 case (H2 b Inb).
 intros x (Hx,Rx).
 replace a with x.
 auto.
 case Rba.
 intros Ha (Hb,Hab).
  
 apply ( H3   _ _ b   Hx Ha Rx Hab).
 red;intros b a a';intros.
 red in H5; decompose [and] H5.
 red in H6; decompose [and] H6.
 clear H5 H6.
 apply (H3 a a' b);auto.
 red; intros a;intros.
 case (H a);auto.
 intros b Rab; exists b.
 split.
 apply H0 with a;auto.
 red;repeat split;auto.
 apply H0 with a.
 auto.
 auto.
 
 red;intros b b' a.
 intros.
 apply (H1 a); auto. 
 case H6;auto.
 case H6;tauto.
 case H7;tauto.
Qed.

 
(** Conversion from a relational definition : A->B->Prop into a partial
    function of type A->B *)

 Section rel_to_fun.
 Variables ( A B : Type).
 Variable inhB :  inhabited B.
 Variables (DA : Ensemble A) (DB : Ensemble B)
           (Rf : A -> B -> Prop).
            
  
 Hypothesis ff : (rel_functional  DA  Rf).
 Hypothesis fd : (rel_domain  DA  Rf).
 Hypothesis fc : rel_codomain DA DB Rf.

 
Definition r2i := iota_fun inhB DA DB Rf.

 Lemma r2i_ok : fun_rel_same  DA   r2i Rf.
 Proof.

 unfold fun_rel_same.
 intros; pattern (r2i a).
 unfold r2i.
  split.
 intro.
 rewrite (iota_fun_rw inhB DA DA DB  Rf a H b).
 auto.
 exists b;split;auto.
 split;auto.
  red in fc. apply fc with a. auto. auto.
 intros. eapply ff with a. auto. auto. tauto. auto. auto. eapply fc;eauto.
intro. 

rewrite <- H0.
eapply iota_fun_ind; eauto.
red in fd, ff.
destruct (fd  H).
exists x;split;auto.
split;auto.
red in fc. 
apply fc with a;auto.
destruct 1;intros; eapply ff;eauto.
destruct 1.
tauto.
Qed.

 Definition the_fun (Hb :  inhabited B)(DA : A->Prop)(R:A->B->Prop)
       (pi : forall a, DA a -> ex (unique (R a))) (a:A):=
   (iota Hb (fun (b:B) => DA a /\ R a b)).

 Theorem the_fun_rw : forall (Hb :  inhabited B)(P: A-> Prop)(R:A->B->Prop)
(pi : forall a, P a -> ex (unique ( R a))) (a:A)(b0:B),
      P a -> R a b0 -> b0= (the_fun Hb P R pi a) . 
Proof.
 intros Hb P R pi a b1 H H0. 
 unfold the_fun.
 pattern (iota  Hb (fun b1 : B => P a /\ R a b1)). 
 iota_elim.
 firstorder.
 case (pi a H).
 intros.
destruct H2.
 symmetry;apply H3.
 auto.
Qed.

Theorem the_fun_e : forall (Hb :  inhabited B)(P: A-> Prop)(R:A->B->Prop)
(pi : forall a, P a -> ex (unique (R a))) (a:A),
      P a -> R a (the_fun Hb P R pi a) . 
Proof.
 intros Hb P R pi a H. 
 unfold the_fun.
 pattern (iota  Hb (fun b0 : B => P a /\ R a b0)). 
 iota_elim.
 case (pi a H).
 intros x (Hx,H'x).
 firstorder.
 firstorder.
Qed.

Theorem the_fun_ind : forall (Hb :  inhabited B)(P: A-> Prop)(Q R:A->B->Prop)
(pi : forall a, P a -> ex (unique (R a))) (a x:A),
      P a -> a = x -> 
(forall b0, P a -> unique (R a) b0 -> Q a b0) ->
 Q x (the_fun Hb P R pi a) . 
Proof.
 intros.
 subst x.
 apply H1. auto.
 case (pi a H).
 destruct 1.
 split.
 eapply the_fun_e.
 auto.
 intros.
 replace x' with x.
 symmetry;apply (the_fun_rw Hb P R pi a);auto.
 auto.
Qed.


 Definition r2f : A -> B.

 refine (the_fun inhB DA Rf _).
 intros a Ha; case (fd Ha);intros x Hx.
 exists x;split;auto.
 intros;eapply ff;eauto.
 
 Defined.



 
End rel_to_fun.



Section inversion_of_bijection.
 Variables (A B : Type)
           (inhA :  inhabited A)
            (inhB :  inhabited B)
           (DA : Ensemble A)
           (DB : Ensemble B)
           (f : A -> B).
 
 Let inv_spec := fun y x => In DA x /\ In DB y /\ f x = y.
 

 Definition inv_fun := r2i  inhA DB DA inv_spec. 

 Hypothesis f_b : fun_bijection DA DB f. 

 Lemma inv_spec_ex : forall b, In DB b -> exists a, inv_spec b a.
 Proof.
   intros.
   case f_b.
   intros.
   case (H1 b H).
 intros x (Hx, fxb).
 unfold inv_spec.
 exists x;auto.
 Qed.

Lemma inv_spec_func : forall b, In DB b ->
                 forall a a', inv_spec b a -> inv_spec b a' -> a=a'. 
Proof.
  intros.
  case f_b;intros.
  red in H4.
  apply H4.
  case H0;auto.
  case H1;auto.
  case H0;case H1;intros.
  transitivity b;intuition.
Qed.



Lemma inv_compose :  
                    forall x, DA x -> inv_fun (f x) = x.
Proof.
 intros.
 unfold inv_fun, r2i.
 pattern (f x), (iota_fun inhA DB DA inv_spec (f x)).
 eapply (iota_fun_ind inhA DB DA).
 destruct f_b.
now apply H0.
 reflexivity.
 exists x.
split.
destruct f_b.
split;auto.
red.
split;auto.
intros.
destruct f_b.
apply H3.
auto.
 tauto.
destruct H0.
destruct H0.
destruct H5;auto.
destruct f_b.
apply H0;auto.
intros.
destruct H0.
apply H1.
split;auto.
split;auto.
split;auto.
 apply f_b.
 auto.
Qed.


Lemma inv_composeR :  
                    forall b, DB b -> f (inv_fun b) = b.
Proof.
 intros.
 unfold inv_fun,r2i.
 pattern  b, (iota_fun inhA DB DA inv_spec b);eapply iota_fun_ind;eauto.
destruct f_b.
case (H1 b H); intros x (H',H5).
exists x;auto.
red;intuition.
red.
auto.
apply H2;auto.
destruct H4.
subst. destruct H4;auto. 
destruct 1.
destruct H0.
destruct H0.  tauto. 
Qed.


Lemma inv_fun_bij : fun_bijection DB DA inv_fun.
 destruct f_b; split.
 red.
 intros b Hb.
unfold inv_fun,r2i.
pattern  b, (iota_fun inhA DB DA inv_spec b);eapply iota_fun_ind;eauto.
destruct (H0 b);auto.
  exists x;split;auto.
  unfold inv_spec.
  repeat split; auto.
 tauto.
 tauto.

 tauto.
  destruct 1.
  apply H1; auto.
 tauto.
 destruct H3.
   destruct H2, H5; congruence.
 destruct 1.
  tauto.
   red.
   intros a Ha; exists (f a);split;auto.
   apply inv_compose;auto.
   red.
   intros b b' Hb Hb' H2.
   rewrite <- (inv_composeR  Hb).
    rewrite <- (inv_composeR  Hb').
    now rewrite H2.
Qed.


 
End inversion_of_bijection.











