
(*  Pierre Casteran 
    LaBRI, Universit Bordeaux 1
*)

(* an injection from a notation for epsilon0 into denumerable ordinals *)




Require Import Epsilon0.
Require Import Schutte.
Require Import Plus.
Require Import AP.
Require Import CNF.


Fixpoint inject (t:T1) : OT :=
 match t with Epsilon0.zero => zero
            | Epsilon0.ocons a n b => mult_Sn (phi0 (inject a)) n  + inject b
 end.


Lemma finite_finite : forall n, inject (FS n) = F (S n).
Proof.
 induction n;simpl.
 -  rewrite phi0_zero.
    rewrite alpha_plus_zero; auto with schutte.
 - clear IHn.  induction n; simpl ; auto.
    + repeat rewrite phi0_zero.
      rewrite alpha_plus_zero; auto with schutte.
      * rewrite <- succ_is_plus_1.
             f_equal. 
             auto with schutte. 
    + rewrite alpha_plus_zero. 
     *  rewrite alpha_plus_zero in IHn.
        rewrite IHn. 
        replace (phi0 zero) with (F 1).
        rewrite <- succ_is_plus_1; auto with schutte.
        symmetry; auto with schutte. 
        apply phi0_zero. 
        auto with schutte. 
     * auto with schutte.
Qed.



Lemma zero_zero : inject Epsilon0.zero = zero.
Proof. reflexivity. Qed.



Lemma omega_omega : inject Epsilon0.omega = omega.
Proof.
 simpl.
 repeat rewrite alpha_plus_zero;try ordi.
 rewrite phi0_zero.
 generalize omega_second_AP.
 destruct 1.
 decompose [and] H0.
 clear H0.
 repeat rewrite zero_plus_alpha;auto with schutte.
 red in H1.
 unfold Ensembles.In in H2.
 tricho (phi0 (F 1)) omega H4.
 2:auto.
 case (lt_omega_finite  H4).
 intros.
 generalize  (AP_finite_eq_one  x).
 intro.
 rewrite <- H0 in H3.
 assert (x=1).
 apply H3.
 apply AP_phi0;auto with schutte.
 subst x.
 rewrite <- phi0_zero in H0.
 absurd (phi0 zero = zero).
 red.
 intro.
 simpl in H5.
 absurd (zero < phi0 zero).
 rewrite H5;apply lt_irr.
 apply phi0_positive;auto with schutte.
 apply phi0_inj;auto with schutte.
 decompose [and] H1; clear H1.
 case (AP_to_phi0 H3).
 intros beta (Hbeta,eqbta).
 rewrite eqbta in H4.
 assert (beta < F 1).
 apply phi0_mono_R ;auto with schutte.
 case (finite_lt_inv  _ H1).
 intros x (Hx,Hb).
 inversion Hx.
 subst x;subst beta.
 simpl in eqbta.
 rewrite phi0_zero in eqbta. 
 rewrite eqbta in H0.
 case (@lt_irr _ H0).
 inversion H6.
Qed.



Lemma ordinal_inject : forall alpha, nf alpha -> ordinal (inject alpha).
Proof.
  induction alpha.
  simpl.
  auto with schutte.
  simpl.
  intros.
  apply ordinal_plus.
  apply ordinal_mult_Sn.
  apply ordinal_phi0.
  apply IHalpha1.
  eapply nf_inv1;eauto.  
  apply IHalpha2.
  eapply nf_inv2;eauto.
Qed.

Hint Resolve ordinal_inject : schutte.

Lemma phi0_mult_lt_phi0 : forall alpha beta, alpha < beta ->
                                             forall n, mult_Sn (phi0 alpha) n  < phi0 beta.
Proof.
  induction n;simpl;auto.
  apply phi0_mono;auto.
  apply AP_plus_closed.
  apply AP_phi0.
  ordi.
  auto.
  apply phi0_mono;auto.
Qed.

Lemma phi0_mult_plus_lt_phi0 :
  forall alpha beta n gamma, alpha < gamma -> beta < phi0 gamma ->
                             mult_Sn (phi0 alpha) n  + beta < phi0 gamma.
Proof.
  intros;apply AP_plus_closed.
  apply AP_phi0;ordi.
  apply  phi0_mult_lt_phi0;auto.
  auto.
Qed.

Lemma phi0_mult_plus_lt_phi0R :
  forall alpha beta n gamma, ordinal alpha -> ordinal beta -> ordinal gamma ->
                             mult_Sn (phi0 alpha) n  + beta < phi0 gamma ->
                             alpha < gamma /\ beta < phi0 gamma.

Proof.
  split.
  assert (phi0 alpha < phi0 gamma).
  eapply le_lt_trans.
  2:eexact H2.
  apply le_trans with  (mult_Sn (phi0 alpha) n).
  clear H2; induction n.
  simpl;auto with schutte.
  simpl.
  apply le_plus_r.
  auto with schutte.
  auto with schutte.
  apply le_plus_l;auto with schutte.
  apply phi0_mono_R;auto.
  eapply le_lt_trans.
  2:eexact H2.
  apply le_plus_r;auto with schutte.
Qed.

 
Lemma zero_lt : forall alpha n beta, ordinal alpha -> ordinal beta ->
                zero < mult_Sn (phi0 alpha) n + beta.
Proof.
 intros alpha n beta Halpha Hbeta.
 apply lt_le_trans with (phi0 alpha).
 apply phi0_positive;auto.
 apply le_trans with ( mult_Sn (phi0 alpha) n).
 elim n; simpl;auto with schutte.
 intros.
 apply le_plus_r;auto.
 auto with schutte.
 auto with schutte.
 apply  le_plus_l;auto with schutte.
Qed.
 

Lemma head_lt :  forall a a' n n' b b', a  < a' -> b < phi0 a'  ->
                                         ordinal b' ->
                           mult_Sn (phi0 a) n + b < mult_Sn (phi0 a') n' + b'.
Proof.
 intros.
 apply lt_le_trans with (phi0 a').
 apply phi0_mult_plus_lt_phi0;auto.
 apply le_trans with (mult_Sn (phi0 a') n').
 elim n';simpl;auto.
 assert (ordinal a').
 ordi.
 auto with schutte.
 assert (ordinal a');[ordi|idtac].
 intros;apply le_plus_r;auto with schutte.
 intros;apply le_plus_l;auto with schutte.
assert (ordinal a');[ordi|auto with schutte].
Qed.


Lemma coeff_lt : forall a  n n' b b', ordinal a -> b < phi0 a  ->
                                         ordinal b' -> (n < n')%nat ->
                           mult_Sn (phi0 a) n + b < mult_Sn (phi0 a) n' + b'.
Proof.
 intros.
 apply lt_le_trans with (mult_Sn (phi0 a) n').
 apply lt_le_trans with (mult_Sn (phi0 a) n + phi0 a).
 apply plus_mono_r.
 auto.
 auto with schutte.
  auto.
 apply mult_Sn_mono3.
  apply  phi0_positive; auto. auto.
  apply le_plus_l. 
 auto. 
 
 apply ordinal_mult_Sn; apply ordinal_phi0.
 auto.
 auto.
 Qed. 


Lemma lt_lt_0 : forall alpha,
                  nf alpha ->
                  forall beta gamma, 
                    Epsilon0.lt  beta gamma -> 
                    Epsilon0.lt gamma alpha ->
                    nf beta -> nf gamma -> inject beta < inject gamma.
Proof. 
 intros alpha . 
 transfinite_induction alpha.

 intros x Indx Nx. 
 induction beta; destruct gamma.
 {  intros.
 T1_inversion H.
 }
 { intros.
   simpl.
   apply lt_le_trans with (phi0 (inject gamma1)).
   apply phi0_positive;auto with schutte.
    apply ordinal_inject; auto with schutte.
   eauto with T1.
   eapply le_trans. 
    2:eapply le_plus_l; auto with schutte.
    apply le_a_mult_Sn_a; auto with schutte.
  apply ordinal_phi0.
   apply ordinal_inject; auto with schutte.
   eauto with T1.
  apply ordinal_mult_Sn; apply ordinal_phi0.
   apply ordinal_inject; auto with schutte.
   eauto with T1.
apply ordinal_inject; auto with schutte.
eauto with T1.
 }
 intros.
  { T1_inversion H. }
intros.
   simpl. 
   destruct (lt_inv H).
   apply head_lt.    
    eapply IHbeta1.
   auto. 
   apply Epsilon0.lt_trans with  (ocons gamma1 n0 gamma2).
   apply Epsilon0.head_lt_ocons.
   auto. 
   eauto with T1.
   eauto with T1.
   
   apply lt_trans with (inject (Epsilon0.phi0 beta1)). 

   eapply IHbeta2.
   apply lt_phi0_phi0.
   apply lt_phi0_intro with n; auto. 
     
   apply  Epsilon0.le_lt_trans with (ocons beta1 n beta2); auto with T1.

   apply le_phi0.
    eapply Epsilon0.lt_trans ; eauto.
   eauto with T1.
  apply nf_phi0.
  eauto with T1.

  simpl. 

   rewrite alpha_plus_zero.
  apply phi0_mono.
   apply IHbeta1; auto. 
   apply Epsilon0.lt_trans with (ocons gamma1 n0 gamma2); auto with T1.
   apply head_lt_ocons.
   eauto with T1.
   eauto with T1.
  apply ordinal_phi0. 
  apply ordinal_inject. eauto with T1.
  apply ordinal_inject. eauto with T1.
  eauto with T1.
  decompose [or and] H3.
  subst.
  apply coeff_lt. apply ordinal_inject; auto. 
  eauto with T1.

  replace  (phi0 (inject gamma1)) with (inject (Epsilon0.phi0 gamma1)).
apply IHbeta2.
apply lt_phi0_phi0.
eapply lt_phi0_intro; eauto.
apply Epsilon0.le_lt_trans with (ocons gamma1 n0 gamma2); auto. 

destruct n0.

apply le_tail; auto. 
auto with T1.
apply lt_le_incl.
apply Epsilon0.coeff_lt. auto with arith.
eauto with T1.
eapply nf_phi0. eauto with T1.
cbn.

rewrite alpha_plus_zero; eauto with T1.
eapply ordinal_phi0. apply ordinal_inject. auto.
eauto with T1.
apply ordinal_inject. auto.
eauto with T1.
auto. 
subst. 
apply plus_mono_r.
apply ordinal_mult_Sn; apply ordinal_phi0; apply ordinal_inject; auto.
eauto with T1.
apply IHbeta2; eauto with schutte T1.
eapply Epsilon0.lt_trans.
2: eapply H0. auto with T1. apply tail_lt_ocons. auto.
Qed. 


Lemma lt_lt : forall beta gamma, Epsilon0.lt  beta gamma -> 
                               nf beta -> nf gamma -> 
                  inject beta < inject gamma.
Proof.  
 intros.
 apply lt_lt_0 with (Epsilon0.succ gamma);auto.
 apply succ_nf;auto.
 apply Epsilon0.lt_succ;auto.
Qed.

Lemma lt_ltR : forall beta gamma, 
                 nf beta -> nf gamma -> 
                 inject beta < inject gamma -> 
                 (beta < gamma)%cnf.
Proof.  
  intros.
  destruct (lt_eq_lt_dec beta gamma) as [[H2 | H2] | H2].
  auto. 
  now split.  
  subst ;  case (lt_irr  H1).
  destruct (@lt_irr (inject beta)).
  eapply lt_trans with (inject gamma); auto.
  now apply lt_lt.
Qed.




