(* Pierre Castran, LaBRI, Universit Bordeaux 1 *)


Require Import Schutte.
Require Import Ensembles.
Require Import Well_Orders.
Require Import Denumerable.
Require Import PartialFun.
Import Classical.
Import MoreIota.
Import Epsilon.

Set Implicit Arguments.

Definition O_segment (A: Ensemble OT) :=
  Included A ordinal /\
  forall a b, A a -> b < a -> A b.

Definition proper_O_segment (A:OT->Prop) := O_segment A /\ 
                                            ~ (Same_set A ordinal).

(* To do : use vocabulary : surjective, etc. *)

Definition ordering_function (f : OT -> OT)(A B : Ensemble OT) :=
 O_segment A /\
 (Included B ordinal) /\
 (forall a, In A a -> In B (f a)) /\
 (forall b, B b -> exists a, In A a /\ f a = b) /\
 forall a b  , In A a -> In A b -> a < b ->  f a < f b.

Definition ordering_segment (A B : Ensemble OT) :=
  exists f : OT -> OT, ordering_function f A B.

Definition fun_eq (f g : OT -> OT )(A : OT -> Prop) :=
 forall a, In A a -> f a = g a. 

Definition fun_eq_gen (f g : OT -> OT)(A B : OT -> Prop) :=
  Same_set A B /\ forall a, In A a -> f a = g a. 

Definition fun_restrict (f g : OT -> OT)(A B : OT -> Prop) :=
  Included A B /\ forall a, In A a -> f a = g a.

Definition proper_segment_of (B : Ensemble OT)(beta : OT): Ensemble OT  :=
  fun alpha => B alpha /\ alpha < beta /\ B beta.

Lemma inh_Ordinal_sets : inhabited (Ensemble OT).
 Proof.  split; apply Empty_set.    Qed.

Instance Inh_OSets : InH (Ensemble OT).
Proof. split. exact inh_Ordinal_sets. Qed.
 
Definition the_ordering_segment {B}:= the  (fun x => ordering_segment x B).

(* we use epsilon, since we don't use extensionaliy for comparing functions *)
(* shouldn't we ? *)


 Lemma inh_OT_OT : inhabited (OT -> OT).
 Proof. split;  exact (fun (alpha:OT) => alpha). Qed.

 Instance Inh_OT_OT : InH (OT -> OT).
 Proof. split;  apply inh_OT_OT. Qed. 

Definition the_ordering_function {B} := 
   some  (fun f => ordering_function f (@the_ordering_segment B) B).

Definition  normal (f : OT -> OT)(B : Ensemble OT): Prop :=
 ordering_function f ordinal B /\ continuous f ordinal B.
 
Hint Resolve  inh_OT_OT   inh_Ordinal_sets  : schutte.

(* O_segments *)


Lemma ordinal_O_segment : O_segment ordinal.
Proof with eauto with schutte.
 split ...
Qed.


Lemma members_proper :
  forall a, ordinal a -> 
  proper_O_segment (members a).
Proof with eauto with schutte.
  split.
 -  split.
   + destruct 1;auto.
   +  destruct 1;  split ...
 -  red;destruct 1 as  [H0 H1];  absurd (succ a < a).
   +  intro; case (le_not_gt (a:=a) (b:=succ a)) ...
   + case (H1 (succ a)) ...
Qed.


Lemma proper_members : forall A, proper_O_segment A -> exists a, 
                                        ordinal a /\ 
                                        Same_set A (members a).
Proof with eauto with schutte.
intros A  [H H0]. 
 case (not_all_not_ex _ (fun b => ordinal b /\ ~ A b)).
 - intro H1;  apply H0;split.
   +  unfold Included; case H;auto.
   +  intros x H2; apply NNPP; intro H3.
      apply (H1 x); split;auto.
 -  intros x H1; case (well_order  (fun x => ordinal x /\ ~ A x) x). 
   + tauto.  
   + auto.
   +  intros y Hy.
      case Hy;intros H2 ((H3,H4),H5).
      exists y;split...
      red;  split.
      * intros a H6; tricho a y T ...
        case H;auto.
        intros H7 H8;   apply H7;auto.
        split;auto.
        unfold In; case H;auto.
        intros H7 H8; apply H7;auto.
        subst y; case H4;auto.
        case H4; case H;intros.
        apply (H8 a);auto.
      * 
        red; unfold In;intros; apply NNPP.
        case Hy;  intros _ ((H7,H8),H9); intro H10.
        case (H9 x0 ).
        { case H6;  auto. }
        { split;auto.   
          case H6;auto.
        }
        { intro;subst y.
          inversion H6.
          case (@lt_irr x0);auto.  
        }
        intro; case (@lt_irr x0); apply lt_trans with y ...
        case H6;eauto with schutte.
Qed.


Lemma denumerable_segment_proper : forall A : Ensemble OT,
           O_segment A -> denumerable A -> proper_O_segment A.
Proof.
 intros A H H0; split;[auto|idtac].
 intro H1; generalize (Extensionality_Ensembles _ _ _ H1).
 intros; subst A ; now case Non_denum.
Qed.

Lemma ordering_function_ordA : forall f A B, ordering_function f A B ->
                                             Included A ordinal.
Proof.
 destruct 1; case H;auto.
Qed.

Hint Resolve ordering_function_ordA : schutte.

Lemma ordering_function_ordB : forall f A B, ordering_function f A B ->
                                Included B ordinal.
Proof.  destruct 1 ; tauto. Qed.

Hint Resolve ordering_function_ordB : schutte.

Lemma Inc_elim : forall (U:Type) (A B: Ensemble U),
                   Included A B ->
                   forall a, In A  a ->In B a.
 Proof. eauto. Qed.

Hint Resolve Inc_elim: schutte.



Lemma ordering_function_mono : forall f A B, ordering_function f A B ->
   forall a b, In A a -> In A b -> a <b -> f a < f b.
Proof.  destruct 1 ;  tauto. Qed.

Hint Resolve ordering_function_mono.

Lemma  ordering_function_mono_weak : 
 forall f A B, ordering_function f A B ->
   forall a b, In A a -> In A b -> a <= b -> f a <= f b.
Proof.
 destruct 1.
 decompose [and] H0; clear H0.
 destruct 3.
 case H6;intros; subst b.
 left; auto with schutte.
 right;apply H5;auto.
Qed.

Hint Resolve ordering_function_mono_weak.

Lemma ordering_function_monoR : forall f A B, ordering_function f A B ->
   forall a b, In A a -> In A b -> f a < f b -> a < b.
Proof.
 intros f A B H a b H0 H1; case H ; intros H2 H3 H4. 
 decompose [and] H3; clear H3. tricho a b Ht; auto.
 - eapply (ordering_function_ordA H a H0 ).
 - eapply (ordering_function_ordA H b H1 ).
 - subst b; now destruct (@lt_irr (f a)).
 - destruct (@lt_irr (f a)); apply lt_trans with (f b);auto.
Qed.

Hint Resolve ordering_function_monoR.

Lemma Ordering_bijection : forall f A B, ordering_function f A B ->
                                         fun_bijection A B f.
Proof.
  destruct 1. 
  decompose [and] H0; clear H0.
  split;red;auto; intros.
  tricho a a' H7.
  -  case H;intros H7 _; apply H7;auto.
  -  case H;intros H7 _; apply H7;auto.
  -  generalize (H5 _ _ H0 H4 );
    rewrite H6; intro H8; case (@lt_irr  _ (H8  H7));auto.
  -  auto.
  -  generalize (H5 _ _ H4 H0 H7);
    rewrite H6; intro H8; case (@lt_irr _ H8);auto.
Qed.


Lemma  ordering_function_mono_weakR : 
  forall f A B, ordering_function f A B ->
                forall a b, In A a -> In A b ->  f a <= f b -> a <= b.
Proof with auto with schutte.
  intros f A B H a b H0 H1 H2.
  case H.
  intros H3 H4; decompose [and] H4; clear H4.
  destruct H2 as [[H2 H4]  | H2].
  case (Ordering_bijection H).
  intros H10 H11 H12 ;  case (H12 a b H0 H1 H4)...
  -  left;split...
     now apply H3.
  -  right; eapply ordering_function_monoR; eauto.
Qed.

Hint Resolve ordering_function_mono_weakR.


Lemma ordering_function_seg : forall A B, ordering_segment A B ->
                                          O_segment A.
Proof.   now destruct 1 as [f [H _]]. Qed.

Lemma fun_eq_refl : forall (f : OT -> OT)(A : Ensemble OT),
   fun_eq_gen f f A A.
Proof.  repeat split; auto. Qed.

Lemma empty_ordering : forall B, (forall b, ~ B b) ->
                                 ordering_function (fun o => o)
                                                   (members zero)
                                                   B.
Proof.
  intros B H; split.
  - split.
    + intros x Hx; now inversion Hx.
    +    intros a b H0 H1 ; destruct  H0 as [x H3].   
         destruct (not_lt_zero H0);auto.
  -  repeat split.
     + intros b H0; now destruct (H b).
     + intros x Hx. case Hx. intros x0 _ H1.    now destruct (not_lt_zero H1).
     + intros b Hb;case (H b);auto.
     +  intros a b Hb; case (@not_lt_zero a);  case Hb;auto.
Qed.

Lemma O_segment_ordinal : forall A a, O_segment A -> A a -> ordinal a.
Proof.
 intros A a H; destruct H; intro H1; apply H;auto.
Qed.


Lemma O_segment_lt : forall A a b, O_segment A -> A a -> b < a -> A b.
Proof.
 intros A a b [H H0]  H1 H2; now apply H0 with a.
Qed.

Hint Resolve O_segment_ordinal.

(* Theorem 13.3 of Schutte's book *)

Theorem ordering_le : forall f A B, ordering_function f A B ->
                                    forall alpha, In A alpha -> alpha  <= f alpha.
Proof with auto with schutte.
  intros f A B H alpha H0; generalize H0;  pattern alpha; apply Induction.
  - clear alpha H0; unfold ordering_function in H; decompose [and] H.   
    unfold progressive;   intros alpha H6 H7 H8.
    tricho alpha (f alpha) H' ...
    +  apply H2; auto.
    +   assert (f (f alpha) < f alpha).
      {    apply H5.
           - eapply O_segment_lt;eauto with schutte.
           - unfold In;auto.
           - unfold In;auto.
      }
      case (le_not_gt (a:= f alpha)(b:=f (f alpha)));auto.
      apply H7; eauto with schutte.
      eapply O_segment_lt;eauto with schutte.
  -  red in H; decompose [and]   H; eauto.
Qed.

Section ordering_function_unicity_1.
  
 Variables B A1 A2 : Ensemble OT.

 Variables f1 f2 : OT -> OT.
 Hypothesis O1 : ordering_function f1 A1 B.
 Hypothesis O2 : ordering_function f2 A2 B.

 Remark ord_B : Included B ordinal.
 Proof.   case O1; intuition.  Qed.
 
 Remark SA1 : O_segment A1.
 Proof.   case O1;intuition.  Qed.

 Remark SA2 : O_segment A2.
 Proof.  case O2;intuition.  Qed.
 
 Hint Resolve SA2 SA1 ord_B.
 
 Lemma A1_A2 :forall a, In A1 a -> A2 a /\ f1 a = f2 a.
 Proof with eauto with schutte.
  intros a Ha;  generalize Ha;pattern a; apply Induction.
  {
    clear a Ha ; intros a H H0 Ha ;   assert (A2 a).
    {
      assert (forall khi, khi < a -> A2 khi /\ f1 khi = f2 khi).
      {  intros;  apply H0...
         apply O_segment_lt with a;auto.
      }
     apply NNPP; intro H2.
     assert (forall y, A2 y -> y < a).
     { intros y H3;  tricho y a H4 ...
       subst y; now case H2.
       destruct H2;  apply O_segment_lt with y;auto.
     }
     assert (forall y, A2 y -> f1 y < f1 a).
     {   intros y H4;  case O1;intros H5 H6;  decompose [and] H6.
         apply H11...
         apply O_segment_lt with a;auto.
     }  
    case O2;intros H5 H6; decompose [and] H6; clear H6;  case (H8 (f1 a)).
    - case O1;intros H6 H10; decompose [and] H10;clear H10.
      apply H14;  auto.
    -  destruct 1;  generalize (H4 x H6).
       intro;  case (H0 x)...
       + apply O_segment_lt with a;auto.
       + intros H13 H14.
         rewrite H14 in H12.
         rewrite H10 in H12.
         case (@lt_irr (f1 a));auto.
    }
    {   split; trivial.
        assert (is_least_member ordinal lt
                                (fun x => B x /\ 
                                          forall y, y < a ->
                                                    f1 y <> x)
                          (f1 a)).
        { split.
          -  case O1;intros.
             unfold In; intuition.
             apply H4; auto.
          -  split.
             + split.
               * case O1;intros.
                 intuition.
                 apply H3;auto.
               * intros;  red;intro.
                 case (@lt_irr (f1 a)).
                 pattern (f1 a) at 1; rewrite <- H3.
                 case O1;intros.
                 decompose [and] H5.
                 apply H10;auto with schutte.
                 apply O_segment_lt with a;auto with schutte.
            +  destruct 2.
               case O1;intros.
               decompose [and] H6.
               case (H8 x H3).
               intros c (Hc,H'c).
               subst x.
               tricho (f1 c) (f1 a) H12.
               * apply H7;  auto.
               *   tricho c a H13.
                   eauto with schutte.
                   case (H4 c H13);auto.
                   subst a;auto.
                   right;apply H11;eauto with schutte.
               *   auto.
               *  auto. 
               }
        assert (is_least_member ordinal lt
                                (fun x => B x /\ forall y, y < a ->
                                                           f2 y <> x)
                         (f2 a)).
        {  split.
           - case O2;intros.
             case H4;auto.
             unfold In, Included;intuition.
             apply H7; auto.
           -  split.
              + split.
                case O2;intros.
                intuition.
                apply H4;auto.
                intros; red;intro.
                case (@lt_irr (f2 a)).
                pattern (f2 a) at 1; rewrite <- H4.
                case O2;intros.
                decompose [and] H6.
                apply H11;auto with schutte.
                apply O_segment_lt with a;auto with schutte.
              +   destruct 2.
                  case O2;intros.
                  decompose [and] H7.
                  case (H9 x H4).
                  intros c (Hc,H'c);  subst x.
                  tricho (f2 c) (f2 a) H13.
                  apply H8;  auto.
                  tricho c a H14.
                  eauto with schutte.
                  * case (H5 c H14);  auto.
                  * subst a;auto.
                  *  right;apply H12;eauto with schutte.
                  * auto.
                  * auto.
        }  
        eapply least_member_unicity;eauto.
        apply AX1.
        generalize
          (least_member_of_eq   AX1
                                (X:=fun x :OT  =>
                                      B x /\ 
                                      (forall y , y < a -> f2 y <> x))
                                (Y:=fun x  => B x /\
                                              (forall y , y < a -> f1 y <> x)) ). 
  intros H4;  generalize (H4 (f2 a) (f1 a)).
  clear H4;intros.
  rewrite H4;  auto.
  intros x (H5,H6);  split;auto.
  intros y Hy; replace (f1 y) with (f2 y); auto.
  case (H0 y) ...
  auto.
  eapply O_segment_lt with a;auto.
  intros x (H5,H6);  split;auto.
  intros y Hy;  replace (f2 y) with (f1 y);auto.
  case (H0 y) ...
  eapply O_segment_lt with a;auto.
  }
  }
  apply O_segment_ordinal with A1;auto.

Qed.

 End ordering_function_unicity_1.

Section ordering_function_unicity.
  
 Variables B A1 A2 : Ensemble OT.

 Variables f1 f2 : OT -> OT.
 Hypothesis O1 : ordering_function f1 A1 B.
 Hypothesis O2 : ordering_function f2 A2 B.

 
 Lemma A2_A1 :forall a, A2 a -> A1 a /\ f2 a = f1 a.
 Proof.
  intros; eapply A1_A2.   
  - eapply O2.
  - apply O1.
  - auto.
 Qed.


Theorem ordering_function_unicity  : fun_eq_gen f1 f2 A1 A2.
Proof.
 split.
 - split.
   +  intros x Hx; case (A1_A2 O1 O2 _ Hx);auto.
   +  intros x Hx; destruct (A2_A1 Hx);auto.
 -  intros a Ha; destruct (A1_A2 O1 O2 a Ha);auto.
Qed.

End ordering_function_unicity.

Lemma ordering_function_seg_unicity : forall A1 A2 B, 
           ordering_segment A1 B ->
           ordering_segment A2 B -> A1 = A2.
Proof.
 destruct 1 as [f1 Hf1].
 destruct 1 as [f2 Hf2].
 apply Extensionality_Ensembles.
 assert (H : fun_eq_gen f1 f2 A1 A2).
 { eapply ordering_function_unicity;eauto.
 }
 case H;auto.
Qed.


Lemma proper_segment_ordinals : forall (B : OT -> Prop), 
   Included B ordinal ->
    forall beta, In B beta -> Included (proper_segment_of B beta) ordinal.
Proof.
 intros B  H beta Hbeta alpha Halpha; case Halpha; intros; now apply H.
Qed.


(* we try to build an ordering function, and the associated ordering segment
 of any subset B composed of ordinals *)
    

Lemma proper_of_proper : forall B beta beta',
                           ordinal beta -> In B beta ->
                           In (proper_segment_of B beta) beta' ->
                           proper_segment_of B beta' =
                           proper_segment_of (proper_segment_of B beta) beta'.
Proof with  eauto with schutte.
 intros; apply Extensionality_Ensembles; split.
 - red; destruct 1.
   split. 
   + split;auto.
     destruct H3; split.
     apply lt_trans with beta'...
     case H1;tauto.
     auto.
   + decompose [and] H3;  split;auto.
 - red; destruct 1;auto.
   split.
   +  case H2;auto.
   +  split.
     tauto.  
     destruct  H3.
     destruct H4 ...
Qed.

Section building_ordering_function.
 Variable B : Ensemble OT.
 Hypothesis OrdB : Included  B ordinal.

 Hypothesis H_B : forall beta, In B beta ->
                               exists! A : Ensemble OT,
                                    ordering_segment A
                                             (proper_segment_of B beta).
  
 Section beta_fixed.

 Variable beta : OT.
 Hypothesis beta_B : In B beta.

 Remark O_beta : ordinal beta.
 Proof.    apply OrdB;auto.  Qed.
 

 (** let us build an ordering function for (proper_segment_of B beta) *)
  
 Definition _A := iota inh_Ordinal_sets
                       (fun E => ordering_segment E
                                                  (proper_segment_of B beta)).

 Definition _f := epsilon inh_OT_OT
                        (fun f => 
                          ordering_function f _A
			   (proper_segment_of B beta)).


Lemma of_beta' : ordering_function _f _A (proper_segment_of B beta).
Proof.
  pattern _f; unfold _f; apply epsilon_spec;
  destruct (H_B  beta_B) as [x [H1 H2]].
  case H1; intros x0 H;  exists x0;auto.
  unfold _A; apply iota_ind.
  - exists x; split; auto.
  - intros a H0;  destruct H0;  replace a with x; auto.
Qed.



 Remark Bbeta_denum : denumerable (proper_segment_of B beta).
 Proof.
  apply AX2.
  - apply proper_segment_ordinals;auto.
  -  exists beta;  split;auto.
     apply OrdB;auto.
     destruct 1;tauto.
 Qed.

Hint Resolve  of_beta': schutte.

Remark A_denum : denumerable _A.
Proof.
 eapply denumerable_bij_funR.
 - eapply Ordering_bijection;
   eauto with schutte.
 - apply Bbeta_denum;eauto.
Qed.

Lemma Proper_A : proper_O_segment _A.
Proof.
 apply denumerable_segment_proper.
 - eapply SA1; eauto with schutte.
 - eapply A_denum.
Qed.

Lemma g_def1  : exists g_beta: OT,  ordinal g_beta /\ _A = members g_beta.
Proof.
 generalize (proper_members Proper_A); intros (a,Ha);exists a;split.
 - now destruct Ha. 
 - eapply Extensionality_Ensembles; tauto.
Qed.


Lemma g_unic : forall g_beta g_beta', ordinal g_beta ->
                                      ordinal g_beta' ->
                                      _A = members g_beta ->
                                      _A = members g_beta' ->
                                      g_beta = g_beta'.
Proof.
  intros g_beta g_beta' H H0 H1 H2;  rewrite H2 in H1;clear H2.
  assert (Same_set  (members g_beta') (members g_beta)).
  { rewrite H1; split;auto. }
  case H2; unfold Included, members; intros; apply le_antisym.
 - apply not_gt_le; auto with schutte.
   intro H5; generalize (H4 g_beta'); intros H6; unfold In in H6.
   case (@lt_irr g_beta').
   assert (H7 : Intersection OT ordinal (fun b : OT => b < g_beta) g_beta').
   {  split;auto. }
   generalize  (H6 H7).
   inversion 1; auto.
 - apply not_gt_le; auto with schutte.
   intro H5;  generalize (H3 g_beta);intros H6;
   destruct  (@lt_irr g_beta).
   assert (H7 : Intersection OT ordinal (fun b : OT => b < g_beta') g_beta)
     by (split;auto).
   generalize  (H6 H7);  now inversion 1.
Qed.

Definition g := iota inh_OT (fun o => ordinal o /\ _A = members o).

End beta_fixed.

Lemma g_def : forall beta, In  B beta ->  _A beta = members (g beta ).
Proof.
  intros.
  pattern (g beta); unfold g; apply iota_ind.
  - case (g_def1 H).
    intros a' [Ha' Ha'']; exists a'; split;auto.
    intros x (Hx,H'x);  eapply g_unic;eauto.
  -  now intros x ((Hx,H'x),U).
Qed.


Lemma g_ord : forall beta, In B beta -> ordinal (g beta).
Proof.
  intros beta H; pattern (g beta); unfold g; apply iota_ind.
 -  case (g_def1 H); intros x (Hx,H'x).
    exists x;split;auto.
    + intros a' (Ha',Ha'').
      eapply g_unic;eauto.
 -   destruct 1; tauto.
Qed.

Lemma g_domains : fun_codomain B ordinal g.
Proof g_ord.


Lemma g_lemma : 
  forall beta, In B beta ->
       ordering_function (_f beta) (members (g beta))
                                       (proper_segment_of B beta).
Proof.
 intros beta H; rewrite <- g_def.
 - now   apply of_beta'.
 - assumption.
Qed.

Lemma g_mono : forall beta1 beta2, In B beta1 -> In B  beta2 ->
                                   beta1 < beta2 ->
                                   g beta1 < g beta2.
Proof with eauto with schutte.
 intros beta1 beta2 H H0 H1.
 assert( B2 : fun_bijection (members (g beta2))
                            (proper_segment_of B beta2)
                            (_f beta2)) .
 { apply Ordering_bijection, g_lemma ... }
 assert (B3 : In (proper_segment_of B beta2) beta1).
 { split.  apply H. split; assumption. }
  assert (B4 : exists alpha, ordinal alpha /\ alpha < g beta2 /\ _f beta2 alpha= beta1).
 { destruct B2.
   case (H3 beta1) ...
   intros alpha [H5 H6]; exists alpha.
   split;auto.
   case H5;auto.
   split;auto.
   case H5;auto.
 }
 case B4; intros alpha (Ha1,(Ha2,Ha3)).
 assert (B5 : ordering_function (_f beta2) (members alpha) 
                                (proper_segment_of B beta1)).
 { repeat split ...
   - red;destruct 1;auto.
   -  cbv beta delta[In]...
      apply lt_trans with a;eauto.
      case H2;auto.
   -  red; destruct 1;auto.
   -  case B2;intros.
      clear H4 H5; red in H3.
      + generalize (H3 a); intros H4;  clear H3;  case H4.
        * split.
          case H2;auto.
          red.
          apply lt_trans with alpha.
          case H2;auto.
          auto.
        *  auto.
   -  case (g_lemma H0). 
      intros H3 H4; decompose [and] H4; clear H4.
      replace beta1 with (_f beta2 alpha).
      apply H9;auto.
      +  split.
         * case H2;auto.
         * red;  apply lt_trans with alpha;auto.
           case H2;auto.
      +  split;auto.
      + case H2;auto.
   -  destruct 1.
      case (g_lemma H0).
      intros H4 H5;  decompose [and] H4;  clear H4;  decompose [and] H5.
      clear H9 H7 H5; case (H6 b). 
      + split;auto.
        split;auto.
        eapply lt_trans;eauto.
        case H3;auto.
       +  intros a (Ha,Ha');exists a;split;auto.
          split.
          case Ha;auto.
          red;  tricho  a alpha X.
         *  case Ha;auto.
         *  auto.
         * subst a; rewrite Ha3 in Ha'.
           subst b; case H3;intros H33 _.
           case (@lt_irr _ H33).
         * clear H6; assert (beta1 < b).
           { rewrite <- Ha3; rewrite <- Ha'.
             case (g_lemma H0).
             destruct 1.
             intros.
             decompose [and] H7.
             apply H12;auto.
             split;auto.
           } case (@lt_irr beta1).
           apply lt_trans with b;auto.
           case H3;auto.
   - intros.
     case (g_lemma H0).
     destruct 1.
     intros.
     decompose [and] H7.
     apply H12;auto.
     clear H10 H9 H12 H7; split;auto.
     case H2;auto.
     red.
     apply lt_trans with alpha;auto.
     case H2;auto.
     split;auto.
     case H3;auto.
     red; apply lt_trans with alpha;auto.
     case H3;auto.
 }
 generalize (g_lemma H);intro.
 generalize (ordering_function_unicity B5 H2). 
 destruct 1.
 generalize (Extensionality_Ensembles _ _ _ H3).
 intros; generalize (members_eq Ha1 (g_ord H) H5).
 intro; subst alpha;auto.
Qed.

Lemma L3a : O_segment (image B g).
Proof.
 split.
 -  intro a;destruct 1 as [x [H H0]].
    subst a; apply g_ord;auto.
  -  intros gbeta alpha  (beta,(H1,H2)) H3.
     assert (Halpha : ordinal alpha)  by ordi.
     subst gbeta. 
     generalize (g_lemma H1);intro.
     case H; intros.
     decompose [and] H2;clear H2.
     assert (exists beta0, In (proper_segment_of B beta) beta0 
                      /\ _f beta alpha=beta0).
     {  exists (_f beta alpha); split;auto.
        apply H6;  split;auto.
     }
     case H2;intros beta0 (H9,H10).
     clear H2.
     assert (B5 : ordering_function (_f beta) (members alpha) 
                 (proper_segment_of B beta0)).
     { 
       repeat split;auto with schutte. 
       red; destruct 1;auto.
       cbv beta delta[In];   eauto with schutte.
       red; apply lt_trans with a;eauto.
       case H2;auto.
       destruct 1;auto.
       assert( B2 : fun_bijection (members (g beta))
                                  (proper_segment_of B beta) (_f beta)) .
        { apply Ordering_bijection.
          apply g_lemma; auto.
        }
        case B2;intros H7 H11 H12.
        generalize (H7 a).
        intros H13;  clear H7;  case H13.
        2:auto.
        split.
        case H2;auto.
        red;  apply lt_trans with alpha.
        case H2;auto.
        auto.
        case (g_lemma H1). 
        intros.
        decompose [and] H11.
        clear H11.
        replace beta0 with (_f beta alpha).
        apply H16;auto.
        split.
        case H2;auto.
        red.
        apply lt_trans with alpha;auto.
        case H2;auto.
        split;auto.
        case H2;auto.
        case H9.
        auto.
        destruct 1.
        case (g_lemma H1).
        intros.
        decompose [and] H11.
        clear H11.
        decompose [and] H12.
        case (H13 b). 
        split;auto.
        split;auto.
        apply lt_trans with beta0;auto.
        case H7;auto.
        case H9;auto.
        destruct 2; auto.
        intros a (Ha,Ha');exists a;split;auto.
        split.
        case Ha;auto.
        red; tricho  a alpha X.
        case Ha;auto.
        auto.
        subst a; rewrite H10 in Ha'.
        subst b;case H7; intros H77 _.
        case (@lt_irr _ H77).
        assert (beta0 < b).
        { rewrite <- H10; rewrite <- Ha'.
          case (g_lemma H1).
          destruct 1.
          intros.
          decompose [and] H18.
          apply H23;auto.
          split;auto.
        }
        case (@lt_irr beta0).
        apply lt_trans with b;auto.
        case H7;auto.
        intros; case (g_lemma H1).
        destruct 1.
        intros.
        decompose [and] H14.
        apply H19;auto.
        split;auto.
        case H2;auto.
        red.
        apply lt_trans with alpha;auto.
        case H2;auto.
        split;auto.
        case H7;auto.
        red.
        apply lt_trans with alpha;auto.
        case H7;auto.
       }
     generalize (g_lemma (beta:=beta0));intros.
     case H9;intros.
     generalize (H2 H7).
     intros; generalize (ordering_function_unicity B5 H12). 
     destruct 1.
     generalize (Extensionality_Ensembles _ _ _ H13).
     intros.
     generalize (members_eq (a:=alpha)(b:=g beta0) Halpha (g_ord H7) H15 ).
     intro.
     exists beta0;split;auto.
Qed.


 Lemma g_bij : fun_bijection B (image B g) g.
 split.
 red. 
 intros.
 red;red.
 exists a;split;auto.
 red.
 destruct 1.
 exists x;split;auto.
 case H;auto.
 case H;auto.
 red;intros.
 tricho a a' XXX.
 apply OrdB;auto.
 apply OrdB;auto.
 generalize (g_mono H H0 XXX). 
 intro.
 rewrite H1 in H2.
 case (@lt_irr _ H2).
 auto.
generalize (g_mono H0 H XXX). 
 intro.
rewrite H1 in H2.
 case (@lt_irr _ H2).
 Qed.

 Hint Resolve g_bij : schutte.


Let g_1 := inv_fun inh_OT B (image B g) g.

Lemma g_1_bij : fun_bijection (image B g) B g_1.
Proof.
 unfold  g_1.
 apply inv_fun_bij; auto with schutte.
Qed.

Hint Resolve g_1_bij : schutte.

(* Lemma g_bij : fun_bijection  *)

Lemma g_1_of : ordering_function g_1 (image B g) B.

 split.
 apply L3a.
 repeat split.
 auto.
 destruct 1.
 case H.
 intros.
 rewrite <- H1.
 replace (g_1 (g x)) with x.
 auto.
 symmetry;unfold g_1;eapply inv_compose;eauto.
 auto with schutte.
 auto with schutte.

intros.
 exists (g b).
 split;auto.
 exists b.
 split;auto.
 unfold g_1;eapply inv_compose;eauto.
 auto with schutte.
 auto with schutte.
 intros.

 tricho (g_1 a) (g_1 b) XXX; auto with schutte.
 case g_1_bij; intros.
 generalize (H2 a H).
 intros;apply OrdB;auto.
  case g_1_bij; intros.
generalize (H2 b H0).
 intros;apply OrdB;auto.
 replace a with (g (g_1 a)) in H1.
 replace b with (g (g_1 b)) in H1.
 rewrite XXX in H1.
 case (@lt_irr _ H1).
 unfold g_1;
 apply inv_composeR;
 auto with schutte.
 unfold g_1;
 apply inv_composeR;
 auto with schutte.
 case (le_not_gt (a:=a)(b:= b)).
 auto with schutte.
 replace a with (g (g_1 a));
 replace b with (g (g_1 b)) .
 apply g_mono;auto with schutte.
 case g_1_bij;auto with schutte.
 case g_1_bij;auto with schutte.
 unfold g_1;
 apply inv_composeR;
 auto with schutte.
 unfold g_1;
 apply inv_composeR;
 auto with schutte.
 unfold g_1;
 apply inv_composeR;
 auto with schutte.
Qed.

Lemma image_B_g_seg  : ordering_segment (image B g) B.
Proof.
 exists g_1.
 apply g_1_of.
Qed.

(** corresponds to Lemma 3 of Schtte's chapter :
    It is used twice in the building of ordering function for any subset B of ordinal *)


Lemma L3_u : exists ! AB, ordering_segment AB B.
Proof.
 exists (image B g).
 split.
 apply image_B_g_seg.
 intros ;eapply ordering_function_seg_unicity;eauto.
 apply image_B_g_seg.
Qed.


 
End building_ordering_function.

(*
L3_u
     : forall B : Ensemble OT,
       Included B ordinal ->
       (forall beta : OT,
        In B beta ->
        exists! A : Ensemble OT,
          ordering_segment A (proper_segment_of B beta)) ->
       exists! AB : Ensemble OT, ordering_segment AB B
*)



Section building_ordering_function_by_induction.

 Variable B : Ensemble OT.
 Hypothesis OB : Included  B ON.


 Lemma ordering_segments_of_B :
    forall beta, In B beta ->
                 exists! A : Ensemble OT,
                   ordering_segment A
                                    (proper_segment_of B beta).
  Proof with eauto with schutte.
    intros beta Hbeta; pattern beta; apply Induction.
    - 
 
 (* we must prove that the following predicate is progressive 
===========================
   progressive
     (fun o : OT =>
      exists! A : Ensemble OT,
        ordering_segment A (proper_segment_of B o)) *)

 intros a H H0.

 (*

1 subgoal
  
  B : Ensemble OT
  OB : Included B ordinal
  beta : OT
  Hbeta : In B beta
  a : M ON
  H : ordinal a
  H0 : forall b : M ON,
       ordinal b ->
       b < a ->
       exists! A : Ensemble OT, ordering_segment A (proper_segment_of B b)
  ============================
   exists! A : Ensemble OT, ordering_segment A (proper_segment_of B a)

 *)
 
 apply L3_u.
    +  red; destruct 1;auto.
    +   intros beta0 H1;
      rewrite <- (proper_of_proper (B:=B) (beta:=a) (beta':=beta0))...
      apply H0...
        *  case H1.
           destruct 2;ordi.
        *  destruct H1 ; tauto.
        *  destruct H1; tauto.  
    -  apply OB;assumption.
Qed.


 
 Theorem B_of_u : exists !  AB, ordering_segment AB B.
 Proof.
 apply L3_u.
 auto.
 apply ordering_segments_of_B.
 Qed.




Theorem B_of' : exists ! AB, exists f, ordering_function f AB B.
Proof with eauto.
 case B_of_u;intros AB (f,pi); exists AB.
  split.
  -  case f;intros.
     exists x;auto.
  -  intros.
     case H; intros;case f;intros.
     eapply ordering_function_seg_unicity...
Qed.

 



Lemma the_ordering_function_ok :
 ordering_function (@the_ordering_function B) (@the_ordering_segment B) B.
Proof.
 pattern (@the_ordering_function B).
 apply epsilon_spec. 
  pattern (@the_ordering_segment B).
 apply iota_spec.
 apply B_of_u.
Qed.


 Lemma the_ordering_function_eq : forall A f,
    ordering_function f A B -> fun_eq_gen f (@the_ordering_function  B)
                                          A (@the_ordering_segment B).
 Proof with eauto with schutte.
   intros;eapply ordering_function_unicity...
   apply the_ordering_function_ok.
 Qed.


End building_ordering_function_by_induction.


Lemma ordering_function_ordinal_B : forall A B f,
          ordering_function f A B -> forall a, In A a -> ordinal (f a).
Proof.
  intros.
  case H.
  destruct 2.
  decompose [and] H3.
  generalize (H2 _ (H4 a H0)).
 auto.
Qed.

Lemma of_image : forall f A B, ordering_function f A B ->
                               ordering_function f A (image A f).
Proof.
 intros f A B O; case O; intros.
 decompose [and] H0.
 repeat split;auto.
 case H;auto.
 case H;auto.
 red.
 destruct 1.
 apply H1.
 decompose [and] H4.
 subst x;apply H3.
 auto.
 intros; exists a;auto.
Qed.





Section Th13_5.
 Variables (A B : Ensemble OT).
 Variable f : OT -> OT.
 Hypothesis f_ord : ordering_function f A B.

 Section recto.
 


 Hypothesis f_cont : continuous f A B.


 Section M_fixed.
 Variable M : Ensemble OT.
 Hypothesis inc : Included M B.
 Hypothesis ne : Inhabited _ M.
 Hypothesis den : denumerable M.

 Let U := fun u => In A u /\ In M ( f u).

 Remark fbij : fun_bijection A B f.
  apply Ordering_bijection.
  auto.
 Qed.

 
 Remark restrict : fun_bijection U M f.
 Proof.
   split.
   -  destruct 1;auto.
   - red; case fbij.
     intros  H H0 H1 b H2.
     case (H0 b).
     +   apply inc;auto.
     +  intros x (Hx,H'x); exists x;split;auto.
        red;unfold U; split;auto.
        subst b;auto.
   -  red; destruct 1.
      destruct 1.
 intros H3; case fbij;intros; now  apply H6.
Qed.
 
 Remark Inc_U_A : Included U A.
 Proof. now destruct 1. Qed.
 
 Remark den_U : denumerable U.
 eapply denumerable_bij_funR with OT M f;auto.
 apply restrict.
 Qed.
 
(*
  A stage in the development (thanks to Virgile Prevosto)
 Remark den_U_aux : denumerable U.
 eapply denumerable_bij_funR with OT M f;auto.
 apply restrict.
 Qed.
Definition den_U :=
   let _ := den in den_U_aux.

Opaque den_U.
*)


 Remark inh_U : Inhabited _ U.
 Proof.  
   case ne;intros.
   exists (inv_fun inh_OT A B f x).
   red; unfold U; rewrite inv_composeR.
   -  split;auto.
      generalize (Ordering_bijection f_ord).
      intros.
      generalize (inv_fun_bij inh_OT H0).
      intros.
      destruct H1.
      apply H1.
      apply inc;auto.
   -  apply fbij.
   -  apply inc;auto.
 Qed.

 
 Remark im_U_f : (image U f : Ensemble OT)  = M.
 Proof.
   eapply Extensionality_Ensembles.
   split.
   red;destruct 1.
   unfold U in H.
   case H.
   intros.
   red  in H0.
   case H0;intros.
   subst x;auto.
   red.
   intros.
   exists (inv_fun inh_OT A B f x).
   split.
   red;red. 
   generalize (Ordering_bijection f_ord).
   intros.
   split.
   case (inv_fun_bij inh_OT H0).  
   auto.
   intros.
   destruct H0.
   rewrite inv_composeR; auto with schutte.
   apply fbij.
   apply inc;auto.
   rewrite inv_composeR; auto with schutte.
   apply fbij.
   apply inc;auto.
 Qed.

 
 Lemma sup_M_in_B : In B (|_| M).
 Proof.
  rewrite <- im_U_f.   case f_cont.  intros H [H0 H1].
  rewrite H1; auto.
  -  apply H.
     case H0; intros.
     apply H3;auto.
     + apply Inc_U_A.
     + apply inh_U.
     + apply den_U.
  -  apply Inc_U_A.
  -  apply inh_U.
  - apply den_U.
Qed.

End M_fixed.

 Lemma Th_13_5_1 : closed B.
 Proof.
   split.
   - eauto with schutte. 
   - apply sup_M_in_B.
 Qed.


End recto.

Section verso.
 Hypothesis B_closed : closed B.
 
 Section U_fixed.
 Variable U : Ensemble OT.
 Hypothesis U_non_empty : Inhabited _ U. 
 Hypothesis U_den : denumerable U.
 Hypothesis U_inc_A : Included U A.

 (* apply Virgile's technique ? *)

 Remark R1_aux : denumerable (image U f).
 Proof.
 apply denumerable_bij_fun with OT U f.
 case (Ordering_bijection f_ord).
 intros.
 split.
 intros u u_In_U.
 repeat red. exists u; auto.
 intros u u_In_img.
 case u_In_img.
 intros x Hx.
 exists x; assumption.
 intros u1 u2 u1_In_U u2_In_U Heq.
 apply H1; try apply U_inc_A; assumption.
 assumption.
 Qed.

 Definition R1 := let foo := U_den in R1_aux.
 Opaque R1.


 Remark R2 : In B (|_| (image U f)). 
 Proof.
   case B_closed.
 intros.
 apply H0;auto.
 red;red.
 destruct 1.
 case f_ord. 
 intros _ (_,(H2,_)).
 case H1;intros. 
 subst x;apply H2. 
 apply U_inc_A;auto. 
 case U_non_empty;intros.
 exists (f x);auto.
 red;red.
 exists x;auto.
 apply R1_aux.
Qed.

Remark R3: exists alpha, In A alpha /\ f alpha = |_|  (image U f).
Proof.
 case f_ord.
 intros.
 decompose [and] H0.
 case (H2 (|_|image U f)).
 apply R2.
 intros x; exists x;auto.
Qed.

Let alpha_ : OT :=
  (epsilon inh_OT (fun alpha =>    In A alpha /\ f alpha = sup(image U f))).



Lemma alpha_A :   In A alpha_.
Proof.
  unfold alpha_;  apply epsilon_ind.
  - apply R3. 
  -  tauto.
Qed.

Lemma alpha_sup : f alpha_ = |_| (image U f). 
Proof.
 pattern alpha_; epsilon_elim.
 - apply R3.
 - tauto.
Qed.

Remark R5 : forall khi, In U khi -> f khi <= f alpha_.
Proof.
 intros; rewrite alpha_sup;  apply sup_upper_bound.
 -  apply R1.
 -  intros x H0; destruct H0, f_ord as [H1 [H3 [H4 [H5 H6]]]].
    destruct H0; subst x;  apply H3,  H4,  U_inc_A; auto.
 -  exists khi; tauto.
Qed.


Remark R6 : forall khi, In U khi ->  khi <=  alpha_ .
Proof with eauto with schutte.
 intros khi H; case (R5 H).
 -  left; split.
    + generalize (U_inc_A  H).
      case f_ord (* argh *).
      destruct 1.
      intros.
      red; apply H1; auto.
   +  case (Ordering_bijection f_ord). 
      intros H1 H2 H3; apply H3.
    *  apply U_inc_A;auto.
    * apply alpha_A.
    *  case H0;auto.
 -  right; case f_ord; intros _ H1; decompose [and] H1.
    tricho khi alpha_ H7; auto.
   +  generalize (U_inc_A H); intro H5; case f_ord.
      intros  H11 _; case H11;  intros H7 H8; apply H7...
   + case f_ord; intros  H11 _; case H11; intros; apply H5.
     apply alpha_A;auto.
   +    subst khi; case (@lt_irr _ H0).
   + case (@lt_irr (f alpha_));auto.
     apply lt_trans with (f khi); auto.
     * apply H6; auto.
      apply alpha_A;  auto.
Qed.

Remark R000 : forall a, In A a -> ordinal a.
Proof.
 case f_ord;intros H H0 a H1; destruct H as [H _];  apply H;auto.
Qed.

Hint Resolve R000.


Remark R001 : forall u, In U u -> ordinal u.
Proof.
 intros u Hu; assert (In A u) by  (apply U_inc_A;auto); auto.
Qed.

Hint Resolve R001 alpha_A.
 
Remark R7 : |_| U <= alpha_ .
Proof. 
 apply sup_least_upper_bound; trivial.
 -  intros y H;  destruct f_ord as [H0 H1].
    destruct H0;  apply H0;  now apply U_inc_A.
 -  destruct f_ord as [H H0]; destruct H.
    apply H, alpha_A.
 - intros;now apply R6.
Qed.
 
 
Remark R4 : forall khi, In U khi -> khi <= sup U.
Proof.
 intros;  apply sup_upper_bound;auto.
Qed.

 (* Schutte's remark "hence A is closed " is out of this section *)


Lemma A_closed : In A (sup U).
Proof.
  assert (H: O_segment A) by (eapply SA2;  eauto).
  destruct H as [H0 H1]. 
  case R7.
  - intros [H2 H3]; rewrite H3; apply alpha_A.
  - intros;red; eapply H1 with alpha_ ;auto.
    apply alpha_A.
Qed.


Remark R4' : forall khi, In U khi -> f khi <= f (sup U).
Proof.
 intros; case f_ord;intros. 
 decompose [and] H1.
 clear H1; case (R4 H).
 -  destruct 1.
    subst khi;left;auto with schutte.
 -  right;auto.
    apply H6;auto with schutte.
    apply A_closed;auto.
Qed.

Remark R4'' : |_| (image U f) <= f (sup U).
Proof.
 apply sup_least_upper_bound.
 -  apply R1.
 -  intros; destruct H, f_ord, H1; apply H1; destruct H, H2.
    subst y;apply H2;  apply U_inc_A;auto.
  -  destruct f_ord; decompose   [and] H0;  generalize A_closed;  intros.
     generalize (H3  _ H4); intros; apply H1;     auto.
  - intros; destruct H; destruct H;subst y;
    apply ordering_function_mono_weak with A B; auto.
    +  apply  A_closed.
    +  apply sup_upper_bound;auto.
Qed.


Remark R42 : f(sup U) <= |_| (image U f).
Proof.
 apply le_trans with (f alpha_).
 -  case f_ord;intros.
    decompose [and] H0;clear H0.
    case R7.
    +  destruct 1.
       rewrite H4;auto with schutte.
      left.
      split;auto.
    +  right;auto.
       apply H5;auto.
       apply A_closed;auto.
 -  unfold alpha_; apply epsilon_ind.
    + apply R3.
    + destruct 1; left;  split; auto.
     *  destruct f_ord.
        decompose [and] H2.
        apply H3.
        apply H5;  auto.
Qed.

Lemma f_sup_commutes : f (|_| U) = |_| (image U f).
Proof.
 apply le_antisym.
 -  apply R42.
 -  apply R4''. 
Qed.

End U_fixed.


Lemma Th_13_5_2 : continuous f A B.
Proof.
 split.
 red.
 case f_ord;tauto.
 split. 
 case f_ord;intros.
 case H;auto.
 intros. split;auto.
 intros; eapply A_closed;auto.
 intros;symmetry;apply f_sup_commutes;auto.
Qed.


End verso.




End Th13_5.

(*
Check Th_13_5_1.


Th_13_5_1
     : forall (A B : Ensemble OT) (f : OT -> OT),
       ordering_function f A B -> continuous f A B -> closed B

Th_13_5_2
     : forall (A B : Ensemble OT) (f : OT -> OT),
       ordering_function f A B -> closed B -> continuous f A B


*)





Theorem TH_13_6 : forall B f, Included B ordinal -> 
                              normal f B ->
                              closed B /\ Unbounded B.
Proof.
 destruct 2.
 split.
 -  eapply Th_13_5_1;eauto.
 -  intros x H2;  generalize (Ordering_bijection H0).
    destruct 1 as [H3 H4 H5].
    exists (f (succ x));  split.
    +  apply H3; auto with schutte.
    +  apply le_lt_trans with (f x).
       * eapply ordering_le;eauto.
       *  eapply ordering_function_mono;eauto with schutte.
Qed.



Theorem O_segment_unbounded : forall A:Ensemble OT, O_segment A -> 
                                        Unbounded A ->
                                        A = ordinal.
Proof with eauto with schutte.
 intros A H H0; red in H0; apply Extensionality_Ensembles; split.
 - case H ...
 -  intros x Hx.
    destruct (H0 _ Hx) as [x0 H1]...
    destruct H1,  H...
Qed.


Lemma ordering_unbounded_unbounded : 
  forall A B f, Included B ordinal -> ordering_function f A B ->
                                  (Unbounded B <-> Unbounded A).
Proof with auto with schutte.
  intros A B f H H0;split.
  - intro; apply not_denumerable_unbounded.
    case H0;  intros H2 _;auto with schutte.
    case H2; auto.
    intro H2;  assert (H3 : denumerable B).
    {  apply denumerable_bij_fun with OT A  f;auto.
       auto with schutte.
       apply Ordering_bijection;auto.
    }
    case (denumerable_not_Unbounded H H3);auto.
  - intro H1;  apply not_denumerable_unbounded ...
    intro H2;  assert (H3 : denumerable A).
    { apply denumerable_bij_funR with OT B  f;auto.
      apply Ordering_bijection;auto.
    }
 case (denumerable_not_Unbounded (X:= A));auto.
 case H0;  destruct 1;auto.
Qed.




Theorem TH_13_6R : forall A B f, Included B ordinal -> 
                                 ordering_function f A B ->
                                 closed B -> 
                                 Unbounded B -> 
                                 normal f B.
Proof with auto with schutte.
 intros A B f H H0 H1 H2; assert (A = ordinal).
 {   
   apply O_segment_unbounded.
   -    eapply SA1;  apply H0.
   - destruct (ordering_unbounded_unbounded   H H0); auto.
 }
 split.
  - now  subst A.
  -  apply Th_13_5_2; subst A ;auto.
Qed.


Lemma order_function_least_least : 
 forall B f  , ordering_function f ordinal B ->
     is_least_member ordinal lt B (f zero).
Proof with auto with schutte.
 intros b f H; case H; intros H0 H1; decompose [and] H1; clear H1.
 case H; intros H8 (H9,H10); split.
 -  apply H2, H4...
 -  split.
   +  apply H4 ...
   + intros x H1 H5;  decompose [and] H10.
     case (H12 _ H5); intros x0 (Hx0,H'x0); subst x.
     generalize (ordering_function_mono_weak H (a:=zero) (b:=x0)).
     destruct 1 ...
     destruct H11 ...
Qed.




Lemma O_segment_lt_closed : forall A a b, O_segment A -> 
                                          In A b -> 
                                          a < b -> 
                                          In A a. 
Proof.
 intros A a b H H0 H1; case H; intros H2 H3; now eapply H3 with (2:= H1).
Qed.

 
Arguments the_ordering_function : clear implicits.
Arguments the_ordering_segment : clear implicits.
